/*
Copyright (C) 1996-1997 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

// draw.c -- this is the only file outside the refresh that touches the
// vid buffer

#include "gl_state.h"

byte		*draw_chars;				// 8*8 graphic characters
qpic_t		*draw_disc;
qpic_t		*draw_backtile;

int			translate_texture;
gltexture_t char_texture;

// default colour textures
gltexture_t r_white;
gltexture_t r_gamma;
gltexture_t r_black;
gltexture_t r_red;
gltexture_t r_green;
gltexture_t r_blue;

byte		conback_buffer[sizeof(qpic_t) + sizeof(glpic_t)];
qpic_t		*conback = (qpic_t *)&conback_buffer;

int			texels;

// generic misc blah textures
gltexture_t MapshotTexture;
gltexture_t CrosshairTexture;

void GL_InitTextureCache (void);

// similar to the above but for display lists
GLuint GL_MaxDisplayList = 0;
GLuint GL_MakeDisplayListObjects (GLsizei n)
{
    int		i;
    GLuint	dl;

    // API inconsistency!!!
    dl = glGenLists (n);

    // find and store the max
    for (i = 0; i < n; i++)
    {
        // API inconsistency!!!
        if ((&dl)[i] > GL_MaxDisplayList)
        {
            // API inconsistency!!!
            GL_MaxDisplayList = (&dl)[i];
        }
    }
    return dl;
}


//=============================================================================
/* Support Routines */

typedef struct cachepic_s
{
    char		name[MAX_QPATH];
    qpic_t		pic;
    byte		padding[32];	// for appended glpic
} cachepic_t;

#define	MAX_CACHED_PICS		128
cachepic_t	menu_cachepics[MAX_CACHED_PICS];
int			menu_numcachepics;

byte		menuplyr_pixels[4096];

int		pic_texels;
int		pic_count;

qpic_t *Draw_PicFromWad (char *name)
{
    qpic_t	*p;
    glpic_t	*gl;
    char	texname[128];

    p = W_GetLumpName (name);
    gl = (glpic_t *) p->data;

    // hack.  names in the wad are not necessarily 0 terminated properly.  for some reason,
    // va doesn't work but sprintf does... (i've observed other va weirdness elsewhere, where it
    // corrupts strings sent into it, so my guess is that the function is either fucked or
    // it doesn't behave the way i'd originally expected it to)
    sprintf (texname, "wad/%s", name);

    // load from the cache
    gl->gl_texture = GL_LoadQuakeTexture (NULL, texname, p->data, p->width, p->height, TEX_TEXTURE);
    gl->sl = 0;
    gl->sh = 1;
    gl->tl = 0;
    gl->th = 1;

    return p;
}


/*
================
Draw_CachePic
================
*/
qpic_t	*Draw_CachePic (char *path)
{
    cachepic_t	*pic;
    int			i;
    qpic_t		*dat;
    glpic_t		*gl;

    for (pic=menu_cachepics, i=0 ; i<menu_numcachepics ; pic++, i++)
    {
        if (!strcmp (path, pic->name))
        {
            return &pic->pic;
        }
    }

    if (menu_numcachepics == MAX_CACHED_PICS)
    {
        Sys_Error ("menu_numcachepics == MAX_CACHED_PICS");
    }
    menu_numcachepics++;

    strlcpy_s (pic->name, path);

    // load the pic from disk
    dat = (qpic_t *)COM_LoadTempFile (path);

    if (!dat)
    {
        Sys_Error ("Draw_CachePic: failed to load %s", path);
    }
    SwapPic (dat);

    // HACK HACK HACK --- we need to keep the bytes for
    // the translatable player picture just for the menu
    // configuration dialog
    if (!strcmp (path, "gfx/menuplyr.lmp"))
    {
        memcpy (menuplyr_pixels, dat->data, dat->width*dat->height);
    }
    pic->pic.width = dat->width;
    pic->pic.height = dat->height;

    gl = (glpic_t *)pic->pic.data;

    // load from the cache.  the va here is a hack to prevent messing with the original string pointer
    gl->gl_texture = GL_LoadQuakeTexture (NULL, va ("%s", path), dat->data, dat->width, dat->height, TEX_TEXTURE);
    gl->sl = 0;
    gl->sh = 1;
    gl->tl = 0;
    gl->th = 1;

    return &pic->pic;
}

/*
===============
Draw_Init

Cleaned up the whole conback loading mess
===============
*/
void Draw_Init (void)
{
    int			i;
    GLint		proxparam;
    qboolean	OGL14Lie = false;

    // init the texture structs
    GL_InitTextureCache ();

    // clear down any error flag we may have hanging over
    glGetError ();

    if (GL_State.gl_version >= 1.4)
    {
        // test blend squaring
        glBlendFunc (GL_SRC_COLOR, GL_DST_COLOR);

        if (glGetError () != GL_NO_ERROR)
        {
            OGL14Lie = true;
        }

        // restore correct blendfunc
        glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

        if (OGL14Lie)
        {
            Con_Printf ("\nWARNING: Your %s reported OpenGL 1.4 support.\nIt lied.\n", GL_State.gl_renderer);
        }
    }

    // we have to check for an NV3x chipset seperately cos they pretend they do support it, but they don;t provide hardware acceleration.
    // therefore they may actually pass the proxy texture test but still perform like undying gank.
    if (!GL_State.NP2BigDirtyFib)
    {
        // before doing anything at all, we must do a texture proxy query to see if non-power-of-2 textures are properly supported
        // because some cards will report OpenGL 2.0 and/or the presence of the texture_non_power_of_two extension, but still not
        // support them.  You naughty boys.  I can't decide whether ATI or NVidia are the most evil here, as ATI will at least
        // honestly crash, whereas NVidia delude you into a false sense of security then hit you with a bizarre software emulation
        // mode running at less than 1 FPS.
        glGenTextures (1, (GLuint *)&i);
        glBindTexture (GL_TEXTURE_2D, i);

        // attempt to create a proxy texture with np2 dimensions
        glTexImage2D (GL_PROXY_TEXTURE_2D, 0, GL_RGBA, 72, 48, 0, GL_RGBA, GL_UNSIGNED_BYTE, NULL);

        // assume it's not supported unless we can prove otherwise
        GL_State.NP2BigDirtyFib = true;

        // make sure there are no errors first
        if (glGetError () == GL_NO_ERROR)
        {
            // check proxy parameters
            glGetTexLevelParameteriv (GL_PROXY_TEXTURE_2D, 0, GL_TEXTURE_WIDTH, &proxparam);

            if (proxparam == 72)
            {
                glGetTexLevelParameteriv (GL_PROXY_TEXTURE_2D, 0, GL_TEXTURE_HEIGHT, &proxparam);

                if (proxparam == 48)
                {
                    glGetTexLevelParameteriv (GL_PROXY_TEXTURE_2D, 0, GL_TEXTURE_RED_SIZE, &proxparam);

                    if (proxparam)
                    {
                        glGetTexLevelParameteriv (GL_PROXY_TEXTURE_2D, 0, GL_TEXTURE_GREEN_SIZE, &proxparam);

                        if (proxparam)
                        {
                            glGetTexLevelParameteriv (GL_PROXY_TEXTURE_2D, 0, GL_TEXTURE_BLUE_SIZE, &proxparam);

                            if (proxparam)
                            {
                                // we know it works now
                                GL_State.NP2BigDirtyFib = false;
                            }
                        }
                    }
                }
            }
        }
        glDeleteTextures (1, (const GLuint *)&i);
    }

    // bring up at an invalid resolution initially, we'll set their proper values the first time we run
    // GL_Set2D cos we don't have all the data we need to do so yet.
    // we need to set vid.width to the minimum accepted value so that any writes to the console text
    // buffer are validly done before we set it's actual correct value.
    vid.conwidth = vid.width = 640;
    vid.conheight = vid.height = 0;

    // explicitly set the ortho init to false so that it's reset properly after doing this
    GL_State.OrthoSet = false;

    // load the console background and the charset
    // by hand, because we need to write the version
    // string into the background before turning
    // it into a texture
    draw_chars = W_GetLumpName ("conchars");

    for (i = 0; i < 256 * 64; i++)
    {
        if (draw_chars[i] == 0)
        {
            // set the proper transparent color
            draw_chars[i] = 255;
        }
    }

    // now turn them into textures
    GL_LoadQuakeTexture (&char_texture, "wad/charset", draw_chars, 128, 128, TEX_TEXTURE);

    conback = Draw_CachePic ("gfx/conback.lmp");

    // save a texture slot for translated picture
    GL_MakeTextureObjects (1, (GLuint *)&translate_texture);

    // and for the blah textures
    GL_MakeTextureObjects (1, (GLuint *)&MapshotTexture.gltexnum);

    // get the other pics we need
    draw_disc = Draw_PicFromWad ("disc");
    draw_backtile = Draw_PicFromWad ("backtile");
}



/*
================
Draw_Character

Draws one 8*8 graphics character with 0 being transparent.
It can be clipped to the top of the screen to allow the console to be
smoothly scrolled off.
================
*/
void Draw_Character (int x, int y, int num)
{
    int				row, col;
    float			frow, fcol, size;

    if (num == 32) return;		// space

    num &= 255;

    if (y <= -8) return;			// totally off screen

    row = num>>4;
    col = num&15;

    frow = row*0.0625;
    fcol = col*0.0625;
    size = 0.0625;

    glBindTexture (GL_TEXTURE_2D, char_texture.gltexnum);

    glBegin (GL_QUADS);
    glTexCoord2f (fcol, frow);
    glVertex2f (x, y);
    glTexCoord2f (fcol + size, frow);
    glVertex2f (x+8, y);
    glTexCoord2f (fcol + size, frow + size);
    glVertex2f (x+8, y+8);
    glTexCoord2f (fcol, frow + size);
    glVertex2f (x, y+8);
    glEnd ();
}


void CharBlockBegin (void)
{
    glBindTexture (GL_TEXTURE_2D, char_texture.gltexnum);

    glBegin (GL_QUADS);
}


void CharBlockEnd (void)
{
    glEnd ();
}


// version of the above to work with an already bound tex
void Draw_SingleScaledCharacter (int x, int y, int num, int scale)
{
    int				row, col;
    float			frow, fcol, size;

    // don't draw spaces
    if (num == 32) return;

    // 0 to 255 range
    num &= 255;

    // totally off screen
    if (y <= -8) return;

    row = num >> 4;
    col = num & 15;

    frow = row * 0.0625;
    fcol = col * 0.0625;
    size = 0.0625;

    glTexCoord2f (fcol, frow);
    glVertex2f (x, y);
    glTexCoord2f (fcol + size, frow);
    glVertex2f (x + scale, y);
    glTexCoord2f (fcol + size, frow + size);
    glVertex2f (x + scale, y + scale);
    glTexCoord2f (fcol, frow + size);
    glVertex2f (x, y + scale);
}


void Draw_SingleCharacter (int x, int y, int num)
{
    Draw_SingleScaledCharacter (x, y, num, 8);
}


void Draw_SingleCharacterBig (int x, int y, int num)
{
    Draw_SingleScaledCharacter (x, y, num, 24);
}


/*
================
Draw_String
================
*/
void Draw_ScaledString (int x, int y, char *str, int scale)
{
    CharBlockBegin ();

    while (*str)
    {
        Draw_SingleScaledCharacter (x, y, *str, scale);
        str++;
        x += scale;
    }
    CharBlockEnd ();
}


void Draw_ScaledStringOrange (int x, int y, char *str, int scale)
{
    CharBlockBegin ();

    while (*str)
    {
        Draw_SingleScaledCharacter (x, y, (*str) + 128, scale);
        str++;
        x += scale;
    }
    CharBlockEnd ();
}


void Draw_String (int x, int y, char *str)
{
    Draw_ScaledString (x, y, str, 8);
}


void Draw_StringBig (int x, int y, char *str)
{
    Draw_ScaledString (x, y, str, 24);
}


/*
================
Draw_DebugChar

Draws a single character directly to the upper right corner of the screen.
This is for debugging lockups by drawing different chars in different parts
of the code.
================
*/
void Draw_DebugChar (char num)
{
}

/*
=============
Draw_AlphaPic
=============
*/
void Draw_AlphaPic (int x, int y, qpic_t *pic, float alpha)
{
    glpic_t			*gl;

    gl = (glpic_t *)pic->data;
    glDisable(GL_ALPHA_TEST);
    glEnable (GL_BLEND);
    glColor4f (1,1,1,alpha);
    glBindTexture (GL_TEXTURE_2D, gl->gl_texture->gltexnum);
    glBegin (GL_QUADS);
    glTexCoord2f (gl->sl, gl->tl);
    glVertex2f (x, y);
    glTexCoord2f (gl->sh, gl->tl);
    glVertex2f (x+pic->width, y);
    glTexCoord2f (gl->sh, gl->th);
    glVertex2f (x+pic->width, y+pic->height);
    glTexCoord2f (gl->sl, gl->th);
    glVertex2f (x, y+pic->height);
    glEnd ();
    glColor4f (1,1,1,1);
    glEnable(GL_ALPHA_TEST);
    glDisable (GL_BLEND);
}


/*
=============
Draw_Pic
=============
*/
void Draw_Pic (int x, int y, qpic_t *pic)
{
    glpic_t			*gl;

    gl = (glpic_t *)pic->data;
    glColor4f (1,1,1,1);
    glBindTexture (GL_TEXTURE_2D, gl->gl_texture->gltexnum);
    glBegin (GL_QUADS);
    glTexCoord2f (gl->sl, gl->tl);
    glVertex2f (x, y);
    glTexCoord2f (gl->sh, gl->tl);
    glVertex2f (x+pic->width, y);
    glTexCoord2f (gl->sh, gl->th);
    glVertex2f (x+pic->width, y+pic->height);
    glTexCoord2f (gl->sl, gl->th);
    glVertex2f (x, y+pic->height);
    glEnd ();
}


/*
=============
Draw_TransPic
=============
*/
void Draw_TransPic (int x, int y, qpic_t *pic)
{
    if (x < 0 || (unsigned)(x + pic->width) > vid.width || y < 0 || (unsigned)(y + pic->height) > vid.height) return;
    Draw_Pic (x, y, pic);
}


void Draw_TransPicAlpha (int x, int y, qpic_t *pic, float alpha)
{
    if (x < 0 || (unsigned)(x + pic->width) > vid.width || y < 0 || (unsigned)(y + pic->height) > vid.height) return;
    Draw_AlphaPic (x, y, pic, alpha);
}


/*
=============
Draw_TransPicTranslate

Only used for the player color selection menu
=============
*/
void Draw_TransPicTranslate (int x, int y, qpic_t *pic, byte *translation)
{
    int				v, u, c;
    unsigned		trans[64*64], *dest;
    byte			*src;
    int				p;

    c = pic->width * pic->height;

    dest = trans;

    for (v=0 ; v<64 ; v++, dest += 64)
    {
        src = &menuplyr_pixels[ ((v*pic->height)>>6) *pic->width];

        for (u=0 ; u<64 ; u++)
        {
            p = src[(u*pic->width)>>6];

            if (p == 255)
            {
                dest[u] = 0;	// was = p causing red fringes!!!
            }
            else
            {
                dest[u] =  d_8to24table[translation[p]];
            }
        }
    }
    glBindTexture (GL_TEXTURE_2D, translate_texture);

    GL_UploadTextureToOpenGL (trans, 64, 64, TEX_NOCOMPRESS);

    glColor3f (1,1,1);
    glBegin (GL_QUADS);
    glTexCoord2f (0, 0);
    glVertex2f (x, y);
    glTexCoord2f (1, 0);
    glVertex2f (x+pic->width, y);
    glTexCoord2f (1, 1);
    glVertex2f (x+pic->width, y+pic->height);
    glTexCoord2f (0, 1);
    glVertex2f (x, y+pic->height);
    glEnd ();
}

void Draw_TextBox (int x, int y, int width, int lines)
{
    qpic_t	*p;
    int		cx, cy;
    int		n;

    // draw left side
    cx = x;
    cy = y;

    p = Draw_CachePic ("gfx/box_tl.lmp");
    Draw_TransPic (cx, cy, p);

    p = Draw_CachePic ("gfx/box_ml.lmp");

    for (n = 0; n < lines; n++)
    {
        cy += 8;
        Draw_TransPic (cx, cy, p);
    }

    p = Draw_CachePic ("gfx/box_bl.lmp");
    Draw_TransPic (cx, cy + 8, p);

    // draw middle
    cx += 8;

    while (width > 0)
    {
        cy = y;

        p = Draw_CachePic ("gfx/box_tm.lmp");
        Draw_TransPic (cx, cy, p);

        p = Draw_CachePic ("gfx/box_mm.lmp");

        for (n = 0; n < lines; n++)
        {
            cy += 8;

            if (n == 1) p = Draw_CachePic ("gfx/box_mm2.lmp");

            Draw_TransPic (cx, cy, p);
        }

        p = Draw_CachePic ("gfx/box_bm.lmp");
        Draw_TransPic (cx, cy + 8, p);

        width -= 2;
        cx += 16;
    }

    // draw right side
    cy = y;

    p = Draw_CachePic ("gfx/box_tr.lmp");
    Draw_TransPic (cx, cy, p);

    p = Draw_CachePic ("gfx/box_mr.lmp");

    for (n = 0; n < lines; n++)
    {
        cy += 8;
        Draw_TransPic (cx, cy, p);
    }

    p = Draw_CachePic ("gfx/box_br.lmp");
    Draw_TransPic (cx, cy + 8, p);
}

float XHairS[8] = {0, 0.25, 0.5, 0.75, 0, 0.25, 0.5, 0.75};
float XHairT[8] = {0, 0, 0, 0, 0.5, 0.5, 0.5, 0.5};
float XHairSAdd = 0.25;
float XHairTAdd = 0.5;

void Draw_XHair (int basex, int basey, int size, float r, float g, float b)
{
    // bound it
    if (crosshair.value > 7) crosshair.value = 7;
    if (crosshair.value < 0) crosshair.value = 0;

    glColor3f (r, g, b);

    glDepthMask (GL_FALSE);
    glEnable (GL_BLEND);
    glBlendFunc (GL_SRC_ALPHA, GL_ONE);
    glDisable (GL_ALPHA_TEST);

    glBindTexture (GL_TEXTURE_2D, CrosshairTexture.gltexnum);

    glTexEnvf (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);

    glBegin (GL_QUADS);

    glTexCoord2f (XHairS[(int) crosshair.value], XHairT[(int) crosshair.value]);
    glVertex2f (basex - size / 2, basey - size / 2);

    glTexCoord2f (XHairS[(int) crosshair.value], XHairT[(int) crosshair.value] + XHairTAdd);
    glVertex2f (basex - size / 2, basey + size / 2);

    glTexCoord2f (XHairS[(int) crosshair.value] + XHairSAdd, XHairT[(int) crosshair.value] + XHairTAdd);
    glVertex2f (basex + size / 2, basey + size / 2);

    glTexCoord2f (XHairS[(int) crosshair.value] + XHairSAdd, XHairT[(int) crosshair.value]);
    glVertex2f (basex + size / 2, basey - size / 2);

    glEnd ();

    glColor3f (1, 1, 1);

    glTexEnvf (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);

    glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
    glDisable (GL_BLEND);
    glDepthMask (GL_TRUE);
    glEnable (GL_ALPHA_TEST);
}

void Draw_XHairWrapper (int basex, int basey, int size)
{
    switch ((int) crosshaircolor.value)
    {
    case 1:
        Draw_XHair (basex, basey, size, 1, 0, 0);
        break;

    case 2:
        Draw_XHair (basex, basey, size, 1, 1, 0);
        break;

    case 3:
        Draw_XHair (basex, basey, size, 0, 1, 0);
        break;

    case 4:
        Draw_XHair (basex, basey, size, 0, 1, 1);
        break;

    case 5:
        Draw_XHair (basex, basey, size, 0, 0, 1);
        break;

    case 6:
        Draw_XHair (basex, basey, size, 1, 0, 1);
        break;

    default:
        Draw_XHair (basex, basey, size, 1, 1, 1);
        break;
    }
}

// because this happens during "think time" it's relatively safe to do it slow...
qboolean Draw_MapShot (char *mapshotname, qboolean needreload)
{
#define MAPSHOTRES 256
#define MAPSHOTBUFFER (void *)(MAPSHOTRES * MAPSHOTRES)
    int		wbegin, wend, hbegin, hend;

    wbegin = GL_State.OrthoWidth / 2 - 280;
    wend = GL_State.OrthoWidth / 2 - 24;

    hbegin = 68;
    hend = 260;

    Draw_TextBox (wbegin - 8, hbegin - 8, 32, 24);

    // always bind first cos GL_UploadTextureToOpenGL in TEX_NOGEN mode assumes that you have a valid initial bind
    glBindTexture (GL_TEXTURE_2D, MapshotTexture.gltexnum);

    // don't hammer the poor hd unless the shot has changed!!!
    if (needreload)
    {
        // we could send this through an alternate path using glTexSubImage, but it doesn't happen during
        // gameplay, so the extra overhead doesn't really matter.
        if (!(GL_LoadQuakeTexture(&MapshotTexture, mapshotname, MAPSHOTBUFFER, MAPSHOTRES, MAPSHOTRES, TEX_NOGEN)))
        {
            return true;
        }
    }

    // draw it
    glBegin (GL_QUADS);

    glTexCoord2f (0, 0);
    glVertex2f (wbegin, hbegin);

    glTexCoord2f (1, 0);
    glVertex2f (wend, hbegin);

    glTexCoord2f (1, 1);
    glVertex2f (wend, hend);

    glTexCoord2f (0, 1);
    glVertex2f (wbegin, hend);

    glEnd ();

    if (needreload)
    {
        Con_Printf ("Drawing: %s\n", mapshotname);
    }

    // dont need to reload
    return false;
}


void Draw_ColorBar (int x, int y, int selected)
{
    int i;
    int cx = x;

    glDisable (GL_TEXTURE_2D);

    glBegin (GL_QUADS);

    for (i = 0; i < 14; i++)
    {
        // take the approximate midpoint colour (handle backward ranges)
        glColor3ubv ((byte *) &d_8to24table[i * 16 + (i < 8 ? 8 : 7)]);

        glVertex2f (cx, y);
        glVertex2f (cx + 8, y);
        glVertex2f (cx + 8, y + 8);
        glVertex2f (cx, y + 8);

        cx += 8;
    }

    // draw the highlight rectangle
    glColor3f (1, 1, 1);

    glVertex2f (x + selected * 8 - 1, y - 1);
    glVertex2f (x + selected * 8 + 9, y - 1);
    glVertex2f (x + selected * 8 + 9, y + 9);
    glVertex2f (x + selected * 8 - 1, y + 9);

    // redraw the highlighted color at brighter intensity (handle backward ranges)
    glColor3ubv ((byte *) &d_8to24table[selected * 16 + (selected < 8 ? 11 : 4)]);

    glVertex2f (x + selected * 8, y);
    glVertex2f (x + selected * 8 + 8, y);
    glVertex2f (x + selected * 8 + 8, y + 8);
    glVertex2f (x + selected * 8, y + 8);

    glEnd ();

    glEnable (GL_TEXTURE_2D);
    glColor3f (1, 1, 1);
}


void DrawMenuBackground (qboolean doQuad)
{
    // reuse the Quad display list here...
    int BaseRes;

    // realtime doesn't increment when disconnected so generate our own time for this
    float objrotate = anglemod (50 * Sys_FloatTime ());

    // textbox - we just do this for the border
    Draw_TextBox (0, 0, GL_State.OrthoWidth / 8 - 2, GL_State.OrthoHeight / 8 - 2);

    // black background
    glDisable (GL_TEXTURE_2D);
    glColor3f (0, 0, 0);

    glBegin (GL_QUADS);
    glVertex2f (8, 8);
    glVertex2f (GL_State.OrthoWidth - 8, 8);
    glVertex2f (GL_State.OrthoWidth - 8, GL_State.OrthoHeight - 8);
    glVertex2f (8, GL_State.OrthoHeight - 8);
    glEnd ();

    glEnable (GL_TEXTURE_2D);
    glColor3f (1, 1, 1);

    // take the lowest of width and height and subtract 64 to get a working base
    BaseRes = ((glwidth > glheight) ? glheight : glwidth) - 64;

    // create a viewport of size baseres center screen
    glViewport ((glwidth - BaseRes) / 2, (glheight - BaseRes) / 2, BaseRes, BaseRes);

    glMatrixMode (GL_PROJECTION);
    glLoadIdentity ();
    gluPerspective (90, 1, 4, 4000);

    glMatrixMode (GL_MODELVIEW);
    glLoadIdentity ();

    // put Z going up and X going deep to match Quakes main render
    // Y remains at right to left, Z becomes top to bottom and X becomes near to far
    glRotatef (-90, 1, 0, 0);
    glRotatef (90, 0, 0, 1);

    // clear the depth buffer
    glClear (GL_DEPTH_BUFFER_BIT);

    // default 3D drawing params
    if (gl_cull.value)
    {
        glEnable (GL_CULL_FACE);
        glCullFace (GL_FRONT);
    }
    else
    {
        glDisable (GL_CULL_FACE);
    }
    glDisable(GL_BLEND);
    glDisable(GL_ALPHA_TEST);
    glEnable(GL_DEPTH_TEST);
    glEnable (GL_TEXTURE_2D);

    if (doQuad && HUD_Quad)
    {
        glTranslatef (30, 0, -21);
        glRotatef (objrotate,  0, 0, 1);

        glCallList (HUD_Quad);
    }
    else if (!doQuad)
    {
        // player model.  load and draw this in realtime cos the skin colours may change
        glTranslatef (50, 0, 0);
        glRotatef (objrotate,  0, 0, 1);

        // if this fails to load it just won't draw anything.  i reckon anyway that if the
        // player mdl fails we've got slightly bigger problems than a menu not working...
        Mod_Draw3DSkinnedWidget ("progs/player.mdl", 0, 13);
    }

    // reset to the 2D view
    GL_Set2D ();
}


/*
================
Draw_ConsoleBackground

================
*/
void Draw_ConsoleBackground (int lines)
{
    int y = (vid.height * 3) >> 2;

    if (lines > y)
    {
        // fullscreen
        Draw_Pic (0, lines - vid.height, conback);
    }
    else
    {
        Draw_AlphaPic (0, lines - vid.height, conback, (float)(1.2 * lines)/y);
    }
}


/*
=============
Draw_TileClear

This repeats a 64*64 tile graphic to fill the screen around a sized down
refresh window.
=============
*/
void Draw_TileClear (int x, int y, int w, int h)
{
    glpic_t *backtilegl = (glpic_t *) draw_backtile->data;

    glColor3f (1,1,1);
    glBindTexture (GL_TEXTURE_2D, backtilegl->gl_texture->gltexnum); //(*(int *)draw_backtile->data);
    glBegin (GL_QUADS);
    glTexCoord2f (x/64.0, y/64.0);
    glVertex2f (x, y);
    glTexCoord2f ( (x+w)/64.0, y/64.0);
    glVertex2f (x+w, y);
    glTexCoord2f ( (x+w)/64.0, (y+h)/64.0);
    glVertex2f (x+w, y+h);
    glTexCoord2f ( x/64.0, (y+h)/64.0 );
    glVertex2f (x, y+h);
    glEnd ();
}


/*
=============
Draw_Fill

Fills a box of pixels with a single color
=============
*/
void Draw_Fill (int x, int y, int w, int h, int c)
{
    glDisable (GL_TEXTURE_2D);
    glColor3f (host_basepal[c*3]/255.0,	host_basepal[c*3+1]/255.0, host_basepal[c*3+2]/255.0);

    glBegin (GL_QUADS);

    glVertex2f (x,y);
    glVertex2f (x+w, y);
    glVertex2f (x+w, y+h);
    glVertex2f (x, y+h);

    glEnd ();

    glColor3f (1,1,1);
    glEnable (GL_TEXTURE_2D);
}
//=============================================================================

/*
================
Draw_FadeScreen

================
*/
void Draw_FadeScreen (void)
{
    // draw the console background for this stuff
    Draw_ConsoleBackground ((int)9999999999);
}

//=============================================================================

/*
================
GL_Set2D

Setup as if the screen was 320*200
================
*/
void GL_Set2D (void)
{
    glViewport (glx, gly, glwidth, glheight);

    glMatrixMode(GL_PROJECTION);
    glLoadIdentity ();

    if (vid.height == 0 && vid.width == 640)
    {
        GL_State.OrthoSet = false;
    }

    // get the correct scale for the orthographic projection so that it's at the same aspect ratio
    // as the regular display, to make compatible with monitors that have funky resolutions
    if (!GL_State.OrthoSet)
    {
        int		i;
        float	lowest;
        int		orthow, orthoh;

        // get the scales
        float wscale = (float) glwidth / 640.0f;
        float hscale = (float) glheight / 480.0f;

        // take the lowest
        lowest = wscale > hscale ? hscale : wscale;

        // divide by lowest and subtract 3 to make the next test immune
        // to fp errors if they should be already multiples of 8
        orthow = ((glwidth / lowest) - 3.0f);
        orthoh = ((glheight / lowest) - 3.0f);

        // take the nearest nultiple of 8 above the result
        for (i = 0; i < orthow; i += 8);
        orthow = i;
        for (i = 0; i < orthoh; i += 8);
        orthoh = i;

        // save them out
        GL_State.OrthoWidth = orthow;
        GL_State.OrthoHeight = orthoh;

        // now give these their correct values
        // we need to set the w and h of the conback texture too otherwise it will have the default initial value
        conback->width = vid.conwidth = vid.width = GL_State.OrthoWidth;
        conback->height = vid.conheight = vid.height = GL_State.OrthoHeight;

        // so it doesn't happen every time...
        GL_State.OrthoSet = true;
    }
    glOrtho  (0, GL_State.OrthoWidth, GL_State.OrthoHeight, 0, -99999, 99999);

    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity ();

    glDisable (GL_DEPTH_TEST);
    glDisable (GL_CULL_FACE);
    glDisable (GL_BLEND);
    glEnable (GL_ALPHA_TEST);

    glColor4f (1, 1, 1, 1);
}

//====================================================================




