/*
Copyright (C) 1996-1997 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/
// gl_vidnt.c -- NT GL vid component

#include "gl_state.h"
#include "winquake.h"
#include "resource.h"
#include <commctrl.h>

#define MAX_MODE_LIST	600
#define VID_ROW_SIZE	3
#define WARP_WIDTH		320
#define WARP_HEIGHT		200
#define MAXWIDTH		10000
#define MAXHEIGHT		10000
#define BASEWIDTH		320
#define BASEHEIGHT		200

#define MODE_WINDOWED			0
#define NO_MODE					(MODE_WINDOWED - 1)
#define MODE_FULLSCREEN_DEFAULT	(MODE_WINDOWED + 1)

typedef struct
{
    modestate_t	type;
    int			width;
    int			height;
    int			modenum;
    int			dib;
    int			fullscreen;
    int			bpp;
    int			refreshrate;
    int			halfscreen;
    char		modedesc[24];
} vmode_t;

typedef struct
{
    int			width;
    int			height;
} lmode_t;

lmode_t	lowresmodes[] =
{
    {320, 200},
    {320, 240},
    {400, 300},
    {512, 384},
};

qboolean		scr_skipupdate;

static vmode_t	modelist[MAX_MODE_LIST];
static int		nummodes;
static vmode_t	*pcurrentmode;
static vmode_t	badmode;

static DEVMODE	gdevmode;
static qboolean	vid_initialized = false;
static qboolean	windowed, leavecurrentmode;
static qboolean vid_canalttab = false;
static qboolean vid_locked = false;	//johnfitz
static qboolean vid_wassuspended = false;
static int		windowed_mouse;
extern qboolean	mouseactive;  // from in_win.c
static HICON	hIcon;

int				DIBWidth, DIBHeight;
RECT			WindowRect;
DWORD			WindowStyle, ExWindowStyle;

HWND			mainwindow, dibwindow;

viddef_t		vid;				// global video state
int				vid_modenum = NO_MODE;
int				vid_realmode;
int				vid_default = MODE_WINDOWED;
static float	vid_gamma = 1.0f;
static int		windowed_default;
unsigned char	vid_curpal[256*3];
static qboolean fullsbardraw = false;

HGLRC			baseRC;
HDC				maindc;

unsigned short	d_8to16table[256];
unsigned int	d_8to24table[256];

float			gldepthmin, gldepthmax;

modestate_t	modestate = MS_UNINIT;

void VID_MenuDraw (void);
void VID_MenuKey (int key);

HWND WINAPI InitializeWindow (HINSTANCE hInstance, int nCmdShow);
LONG WINAPI MainWndProc (HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam);
void AppActivate(BOOL fActive, BOOL minimize);
char *VID_GetModeDescription (int mode);
void ClearAllStates (void);
void VID_UpdateWindowStatus (void);
void GL_Init (void);

//====================================
int desktop_bpp;				// query this @ startup and when entering the video menu
int vid_bpp;
int vid_current_bpp;			//R00k

void VID_Menu_Init (void);		//johnfitz
void VID_Menu_f (void);			//johnfitz

void GL_SetupState (void);		//johnfitz

//====================================

//johnfitz -- new cvars
cvar_t vid_fullscreen = {"vid_fullscreen", "1", true};
cvar_t vid_width = {"vid_width", "640", true};
cvar_t vid_height = {"vid_height", "480", true};
cvar_t vid_refreshrate = {"vid_refreshrate", "60", true};
//johnfitz

cvar_t		vid_mode = {"vid_mode","0", false};
// Note that 0 is MODE_WINDOWED
cvar_t		_vid_default_mode = {"_vid_default_mode","0", true};
// Note that 3 is MODE_FULLSCREEN_DEFAULT
cvar_t		_vid_default_mode_win = {"_vid_default_mode_win","3", true};
cvar_t		vid_wait = {"vid_wait","0"};
cvar_t		vid_nopageflip = {"vid_nopageflip","0", true};
cvar_t		_vid_wait_override = {"_vid_wait_override", "0", true};
cvar_t		vid_config_x = {"vid_config_x","800", true};
cvar_t		vid_config_y = {"vid_config_y","600", true};
cvar_t		vid_stretch_by_2 = {"vid_stretch_by_2","1", true};
cvar_t		_windowed_mouse = {"_windowed_mouse","1", true};

int			window_center_x, window_center_y, window_x, window_y, window_width, window_height;
RECT		window_rect;

void CenterWindow(HWND hWndCenter, int width, int height, BOOL lefttopjustify)
{
    int     CenterX, CenterY;

    CenterX = (GetSystemMetrics(SM_CXSCREEN) - width) / 2;
    CenterY = (GetSystemMetrics(SM_CYSCREEN) - height) / 2;

    if (CenterX > CenterY * 2)
    {
        CenterX >>= 1;	// dual screens
    }
    CenterX = (CenterX < 0) ? 0: CenterX;
    CenterY = (CenterY < 0) ? 0: CenterY;

    SetWindowPos (hWndCenter, NULL, CenterX, CenterY, 0, 0,	SWP_NOSIZE | SWP_NOZORDER | SWP_SHOWWINDOW | SWP_DRAWFRAME);
}

qboolean VID_SetWindowedMode (int modenum)
{
    HDC				hdc;
    int				lastmodestate, width, height;
    RECT			rect;

    lastmodestate = modestate;

    WindowRect.top = WindowRect.left = 0;

    WindowRect.right = modelist[modenum].width;
    WindowRect.bottom = modelist[modenum].height;

    DIBWidth = modelist[modenum].width;
    DIBHeight = modelist[modenum].height;

    WindowStyle = WS_OVERLAPPED | WS_BORDER | WS_CAPTION | WS_SYSMENU | WS_MINIMIZEBOX;
    ExWindowStyle = 0;

    rect = WindowRect;
    AdjustWindowRectEx(&rect, WindowStyle, FALSE, 0);

    width = rect.right - rect.left;
    height = rect.bottom - rect.top;

    // Create the DIB window
    dibwindow = CreateWindowEx (
                    ExWindowStyle,
                    "Realm",
                    "Realm",
                    WindowStyle,
                    rect.left, rect.top,
                    width,
                    height,
                    NULL,
                    NULL,
                    global_hInstance,
                    NULL);

    if (!dibwindow)
        Sys_Error ("Couldn't create DIB window");

    // Center and show the DIB window
    CenterWindow(dibwindow, WindowRect.right - WindowRect.left, WindowRect.bottom - WindowRect.top, false);

    ShowWindow (dibwindow, SW_SHOWDEFAULT);
    UpdateWindow (dibwindow);

    modestate = MS_WINDOWED;

// because we have set the background brush for the window to NULL
// (to avoid flickering when re-sizing the window on the desktop),
// we clear the window to black when created, otherwise it will be
// empty while Quake starts up.
    hdc = GetDC(dibwindow);
    PatBlt(hdc,0,0,WindowRect.right,WindowRect.bottom,BLACKNESS);
    ReleaseDC(dibwindow, hdc);

    vid.width = modelist[modenum].width;
    vid.height = modelist[modenum].height;
    vid.conwidth = vid.width & 0xFFFFFFF8;
    vid.conheight = vid.conwidth * vid.height / vid.width;

    vid.numpages = 2;

    mainwindow = dibwindow;

    SendMessage (mainwindow, WM_SETICON, (WPARAM)TRUE, (LPARAM)hIcon);
    SendMessage (mainwindow, WM_SETICON, (WPARAM)FALSE, (LPARAM)hIcon);

    return true;
}


qboolean VID_SetFullDIBMode (int modenum)
{
    HDC				hdc;
    int				lastmodestate, width, height;
    RECT			rect;

    if (!leavecurrentmode)
    {
        gdevmode.dmFields = DM_BITSPERPEL | DM_PELSWIDTH | DM_PELSHEIGHT;
        gdevmode.dmBitsPerPel = modelist[modenum].bpp;
        gdevmode.dmPelsWidth = modelist[modenum].width << modelist[modenum].halfscreen;
        gdevmode.dmPelsHeight = modelist[modenum].height;
        gdevmode.dmDisplayFrequency = modelist[modenum].refreshrate;
        gdevmode.dmSize = sizeof (gdevmode);

        if (ChangeDisplaySettings (&gdevmode, CDS_FULLSCREEN) != DISP_CHANGE_SUCCESSFUL)
        {
            Sys_Error ("Couldn't set fullscreen DIB mode");
        }
    }

    lastmodestate = modestate;
    modestate = MS_FULLDIB;

    WindowRect.top = WindowRect.left = 0;

    WindowRect.right = modelist[modenum].width;
    WindowRect.bottom = modelist[modenum].height;

    DIBWidth = modelist[modenum].width;
    DIBHeight = modelist[modenum].height;

    WindowStyle = WS_POPUP;
    ExWindowStyle = 0;

    rect = WindowRect;
    AdjustWindowRectEx(&rect, WindowStyle, FALSE, 0);

    width = rect.right - rect.left;
    height = rect.bottom - rect.top;

    // Create the DIB window
    dibwindow = CreateWindowEx (
                    ExWindowStyle,
                    "Realm",
                    "Realm",
                    WindowStyle,
                    rect.left, rect.top,
                    width,
                    height,
                    NULL,
                    NULL,
                    global_hInstance,
                    NULL);

    if (!dibwindow)
    {
        Sys_Error ("Couldn't create DIB window");
    }
    ShowWindow (dibwindow, SW_SHOWDEFAULT);
    UpdateWindow (dibwindow);

    // Because we have set the background brush for the window to NULL
    // (to avoid flickering when re-sizing the window on the desktop), we
    // clear the window to black when created, otherwise it will be
    // empty while Quake starts up.
    hdc = GetDC(dibwindow);
    PatBlt(hdc,0,0,WindowRect.right,WindowRect.bottom,BLACKNESS);
    ReleaseDC(dibwindow, hdc);

    vid.width = modelist[modenum].width;
    vid.height = modelist[modenum].height;
    vid.conwidth = vid.width & 0xFFFFFFF8;
    vid.conheight = vid.conwidth * vid.height / vid.width;

    vid.numpages = 2;

// needed because we're not getting WM_MOVE messages fullscreen on NT
    window_x = 0;
    window_y = 0;

    mainwindow = dibwindow;

    SendMessage (mainwindow, WM_SETICON, (WPARAM)TRUE, (LPARAM)hIcon);
    SendMessage (mainwindow, WM_SETICON, (WPARAM)FALSE, (LPARAM)hIcon);

    return true;
}


int VID_SetMode (int modenum, unsigned char *palette)
{
    int				original_mode, temp;
    static qboolean	stat;
    MSG				msg;

    if ((windowed && (modenum != 0)) ||	(!windowed && (modenum < 1)) ||	(!windowed && (modenum >= nummodes)))
    {
        Sys_Error ("Bad video mode\n");
    }

// so Con_Printfs don't mess us up by forcing vid and snd updates
    temp = scr_disabled_for_loading;
    scr_disabled_for_loading = true;

    CDAudio_Pause ();

    if (vid_modenum == NO_MODE)
        original_mode = windowed_default;
    else
        original_mode = vid_modenum;

    // Set either the fullscreen or windowed mode
    if (modelist[modenum].type == MS_WINDOWED)
    {
        if (_windowed_mouse.value && key_dest == key_game)
        {
            stat = VID_SetWindowedMode(modenum);
            IN_ActivateMouse ();
            IN_HideMouse ();
        }
        else
        {
            IN_DeactivateMouse ();
            IN_ShowMouse ();
            stat = VID_SetWindowedMode(modenum);
        }
    }
    else if (modelist[modenum].type == MS_FULLDIB)
    {
        stat = VID_SetFullDIBMode(modenum);
        IN_ActivateMouse ();
        IN_HideMouse ();
    }
    else
    {
        Sys_Error ("VID_SetMode: Bad mode type in modelist");
    }

    window_width = DIBWidth;
    window_height = DIBHeight;
    VID_UpdateWindowStatus ();

    CDAudio_Resume ();
    scr_disabled_for_loading = temp;

    if (!stat)
    {
        Sys_Error ("Couldn't set video mode");
    }

// now we try to make sure we get the focus on the mode switch, because
// sometimes in some systems we don't.  We grab the foreground, then
// finish setting up, pump all our messages, and sleep for a little while
// to let messages finish bouncing around the system, then we put
// ourselves at the top of the z order, then grab the foreground again,
// Who knows if it helps, but it probably doesn't hurt
    SetForegroundWindow (mainwindow);
    VID_SetPalette (palette);
    vid_modenum = modenum;
    Cvar_SetValue ("vid_mode", (float)vid_modenum);

    while (PeekMessage (&msg, NULL, 0, 0, PM_REMOVE))
    {
        TranslateMessage (&msg);
        DispatchMessage (&msg);
    }

    Sleep (100);

    SetWindowPos (mainwindow, HWND_TOP, 0, 0, 0, 0,
                  SWP_DRAWFRAME | SWP_NOMOVE | SWP_NOSIZE | SWP_SHOWWINDOW |
                  SWP_NOCOPYBITS);

    SetForegroundWindow (mainwindow);

// fix the leftover Alt from any Alt-Tab or the like that switched us away
    ClearAllStates ();

    VID_SetPalette (palette);

    // force initial r_refdef to set
    vid.recalc_refdef = true;

    //R00k mouse died on mode change
    IN_StartupMouse ();

    return true;
}

/*
================
VID_UpdateWindowStatus
================
*/
void VID_UpdateWindowStatus (void)
{
    window_rect.left = window_x;
    window_rect.top = window_y;
    window_rect.right = window_x + window_width;
    window_rect.bottom = window_y + window_height;
    window_center_x = (window_rect.left + window_rect.right) / 2;
    window_center_y = (window_rect.top + window_rect.bottom) / 2;

    IN_UpdateClipCursor ();
}


//====================================

gl_caps_s	GLCaps;

static char	*VendorCode[10] = {"ARB", "EXT", "NV", "ATI", "SGI", "SGIS", "SGIX", "INTEL", "3DFX", NULL};

// internal capabilities
qboolean OpenGL12;
qboolean OpenGL13;
qboolean OpenGL14;
qboolean OpenGL15;
qboolean OpenGLFrag;
qboolean OpenGLVert;

/*
==================
GL EXTENSION MALARKEY

As a general rule, I don't like the idea of proprietary extensions.  A special case of this rule is that
anything which allows me to rub ATI's nose in it must necessarily be a good thing :D

This will check vendor codes for extensions too...
==================
*/
qboolean GL_CheckExtension (char *ExtensionName, qboolean crash, qboolean silent)
{
    int		i;

    // let us specify extensions directly
    if (ExtensionName[0] == 'G' && ExtensionName[1] == 'L' && ExtensionName[2] == '_')
    {
        // allow direct access to the extension name
        if (strstr (GL_State.gl_extensions, ExtensionName))
        {
            if (!silent)
            {
                Con_Printf ("Found Extension %s.\n", ExtensionName);
            }
            return true;
        }

        if (crash)
        {
            Sys_Error ("Extension %s NOT Found!!!\n\nAborting Realm...", ExtensionName);
        }

        if (!silent)
        {
            Con_Printf ("Extension %s NOT Found.\n", ExtensionName);
        }
        return false;
    }

    // use the fallback mechanism
    for (i = 0; ; i++)
    {
        if (!VendorCode[i]) break;

        if (strstr (GL_State.gl_extensions, va ("GL_%s_%s", VendorCode[i], ExtensionName)))
        {
            if (!silent)
            {
                Con_Printf ("Found Extension GL_%s_%s.\n", VendorCode[i], ExtensionName);
            }
            return true;
        }
    }

    if (crash)
    {
        Sys_Error ("Extension GL_*_%s NOT Found!!!\n\nAborting Realm...", ExtensionName);
    }

    if (!silent)
    {
        Con_Printf ("Extension GL_*_%s NOT Found.\n", ExtensionName);
    }
    return false;
}

/*
================
MHGLGetProcAddress

Implements a work-around for earlier OpenGL versions, by checking for vendor-specific entrypoints that
support the required functions.
================
*/
PROC MHGLGetProcAddress (char *funcname, qboolean crash, qboolean *ver)
{
    int		i;
    static	PROC thePROC;

    // see is the function supported by the core implementation
    if (thePROC = wglGetProcAddress (funcname))
    {
        *ver = (*ver && true);
        return thePROC;
    }

    for (i = 0; ; i++)
    {
        if (!VendorCode[i]) break;

        thePROC = wglGetProcAddress (va ("%s%s", funcname, VendorCode[i]));

        if (thePROC)
        {
            Con_Printf ("%s Unavailable, reverting to %s%s\n", funcname, funcname, VendorCode[i]);
            *ver = (*ver && true);
            return thePROC;
        }
    }

    // oh dear.
    if (crash)
    {
        Sys_Error ("Core OpenGL function %s Unavailable!!!", funcname);
    }
    *ver = (*ver && false);
    return NULL;
}

// gets function prototypes to establish what's available.  I don't query the GL_VERSION as some implementations report
// a version higher than it actually supported (ATI, for example, don't support all entry points for their "OpenGL 2"
// driver.  I don't query the extension string as some drivers have support for trimming the string.
void GL_GetFunctionPointers (void)
{
    // initially all true
    OpenGL12 = true;
    OpenGL13 = true;
    OpenGL14 = true;
    OpenGL15 = true;
    OpenGLFrag = true;
    OpenGLVert = true;

    // Realm doesn't have a 1 TMU render path any more, so...
    glGetIntegerv (GL_MAX_TEXTURE_UNITS, &GLCaps.GL_TextureUnits);

    if (GLCaps.GL_TextureUnits < 2)
    {
        Sys_Error ("Not enough TMUs");
    }

    // opengl 1.2 prototypes
    glBlendColor = (void *) MHGLGetProcAddress ("glBlendColor", true, &OpenGL12);
    glBlendEquation = (void *) MHGLGetProcAddress ("glBlendEquation", true, &OpenGL12);
    glDrawRangeElements = (void *) MHGLGetProcAddress ("glDrawRangeElements", true, &OpenGL12);
    glTexImage3D = (void *) MHGLGetProcAddress ("glTexImage3D", true, &OpenGL12);
    glTexSubImage3D = (void *) MHGLGetProcAddress ("glTexSubImage3D", true, &OpenGL12);
    glCopyTexSubImage3D = (void *) MHGLGetProcAddress ("glCopyTexSubImage3D", true, &OpenGL12);

    // OpenGL 1.3 Prototypes
    glActiveTexture = (void *) MHGLGetProcAddress ("glActiveTexture", true, &OpenGL13);
    glClientActiveTexture = (void *) MHGLGetProcAddress ("glClientActiveTexture", true, &OpenGL13);
    glMultiTexCoord1d = (void *) MHGLGetProcAddress ("glMultiTexCoord1d", true, &OpenGL13);
    glMultiTexCoord1dv = (void *) MHGLGetProcAddress ("glMultiTexCoord1dv", true, &OpenGL13);
    glMultiTexCoord1f = (void *) MHGLGetProcAddress ("glMultiTexCoord1f", true, &OpenGL13);
    glMultiTexCoord1fv = (void *) MHGLGetProcAddress ("glMultiTexCoord1fv", true, &OpenGL13);
    glMultiTexCoord1i = (void *) MHGLGetProcAddress ("glMultiTexCoord1i", true, &OpenGL13);
    glMultiTexCoord1iv = (void *) MHGLGetProcAddress ("glMultiTexCoord1iv", true, &OpenGL13);
    glMultiTexCoord1s = (void *) MHGLGetProcAddress ("glMultiTexCoord1s", true, &OpenGL13);
    glMultiTexCoord1sv = (void *) MHGLGetProcAddress ("glMultiTexCoord1sv", true, &OpenGL13);
    glMultiTexCoord2d = (void *) MHGLGetProcAddress ("glMultiTexCoord2d", true, &OpenGL13);
    glMultiTexCoord2dv = (void *) MHGLGetProcAddress ("glMultiTexCoord2dv", true, &OpenGL13);
    glMultiTexCoord2f = (void *) MHGLGetProcAddress ("glMultiTexCoord2f", true, &OpenGL13);
    glMultiTexCoord2fv = (void *) MHGLGetProcAddress ("glMultiTexCoord2fv", true, &OpenGL13);
    glMultiTexCoord2i = (void *) MHGLGetProcAddress ("glMultiTexCoord2i", true, &OpenGL13);
    glMultiTexCoord2iv = (void *) MHGLGetProcAddress ("glMultiTexCoord2iv", true, &OpenGL13);
    glMultiTexCoord2s = (void *) MHGLGetProcAddress ("glMultiTexCoord2s", true, &OpenGL13);
    glMultiTexCoord2sv = (void *) MHGLGetProcAddress ("glMultiTexCoord2sv", true, &OpenGL13);
    glMultiTexCoord3d = (void *) MHGLGetProcAddress ("glMultiTexCoord3d", true, &OpenGL13);
    glMultiTexCoord3dv = (void *) MHGLGetProcAddress ("glMultiTexCoord3dv", true, &OpenGL13);
    glMultiTexCoord3f = (void *) MHGLGetProcAddress ("glMultiTexCoord3f", true, &OpenGL13);
    glMultiTexCoord3fv = (void *) MHGLGetProcAddress ("glMultiTexCoord3fv", true, &OpenGL13);
    glMultiTexCoord3i = (void *) MHGLGetProcAddress ("glMultiTexCoord3i", true, &OpenGL13);
    glMultiTexCoord3iv = (void *) MHGLGetProcAddress ("glMultiTexCoord3iv", true, &OpenGL13);
    glMultiTexCoord3s = (void *) MHGLGetProcAddress ("glMultiTexCoord3s", true, &OpenGL13);
    glMultiTexCoord3sv = (void *) MHGLGetProcAddress ("glMultiTexCoord3sv", true, &OpenGL13);
    glMultiTexCoord4d = (void *) MHGLGetProcAddress ("glMultiTexCoord4d", true, &OpenGL13);
    glMultiTexCoord4dv = (void *) MHGLGetProcAddress ("glMultiTexCoord4dv", true, &OpenGL13);
    glMultiTexCoord4f = (void *) MHGLGetProcAddress ("glMultiTexCoord4f", true, &OpenGL13);
    glMultiTexCoord4fv = (void *) MHGLGetProcAddress ("glMultiTexCoord4fv", true, &OpenGL13);
    glMultiTexCoord4i = (void *) MHGLGetProcAddress ("glMultiTexCoord4i", true, &OpenGL13);
    glMultiTexCoord4iv = (void *) MHGLGetProcAddress ("glMultiTexCoord4iv", true, &OpenGL13);
    glMultiTexCoord4s = (void *) MHGLGetProcAddress ("glMultiTexCoord4s", true, &OpenGL13);
    glMultiTexCoord4sv = (void *) MHGLGetProcAddress ("glMultiTexCoord4sv", true, &OpenGL13);
    glLoadTransposeMatrixf = (void *) MHGLGetProcAddress ("glLoadTransposeMatrixf", true, &OpenGL13);
    glLoadTransposeMatrixd = (void *) MHGLGetProcAddress ("glLoadTransposeMatrixd", true, &OpenGL13);
    glMultTransposeMatrixf = (void *) MHGLGetProcAddress ("glMultTransposeMatrixf", true, &OpenGL13);
    glMultTransposeMatrixd = (void *) MHGLGetProcAddress ("glMultTransposeMatrixd", true, &OpenGL13);
    glSampleCoverage = (void *) MHGLGetProcAddress ("glSampleCoverage", true, &OpenGL13);
    glCompressedTexImage3D = (void *) MHGLGetProcAddress ("glCompressedTexImage3D", true, &OpenGL13);
    glCompressedTexImage2D = (void *) MHGLGetProcAddress ("glCompressedTexImage2D", true, &OpenGL13);
    glCompressedTexImage1D = (void *) MHGLGetProcAddress ("glCompressedTexImage1D", true, &OpenGL13);
    glCompressedTexSubImage3D = (void *) MHGLGetProcAddress ("glCompressedTexSubImage3D", true, &OpenGL13);
    glCompressedTexSubImage2D = (void *) MHGLGetProcAddress ("glCompressedTexSubImage2D", true, &OpenGL13);
    glCompressedTexSubImage1D = (void *) MHGLGetProcAddress ("glCompressedTexSubImage1D", true, &OpenGL13);
    glGetCompressedTexImage = (void *) MHGLGetProcAddress ("glGetCompressedTexImage", true, &OpenGL13);

    // OpenGL 1.4 Prototypes
    glBlendFuncSeparate = (void *) MHGLGetProcAddress ("glBlendFuncSeparate", true, &OpenGL14);
    glFogCoordf = (void *) MHGLGetProcAddress ("glFogCoordf", true, &OpenGL14);
    glFogCoordfv = (void *) MHGLGetProcAddress ("glFogCoordfv", true, &OpenGL14);
    glFogCoordd = (void *) MHGLGetProcAddress ("glFogCoordd", true, &OpenGL14);
    glFogCoorddv = (void *) MHGLGetProcAddress ("glFogCoorddv", true, &OpenGL14);
    glFogCoordPointer = (void *) MHGLGetProcAddress ("glFogCoordPointer", true, &OpenGL14);
    glMultiDrawArrays = (void *) MHGLGetProcAddress ("glMultiDrawArrays", true, &OpenGL14);
    glMultiDrawElements = (void *) MHGLGetProcAddress ("glMultiDrawElements", true, &OpenGL14);
    glPointParameterf = (void *) MHGLGetProcAddress ("glPointParameterf", true, &OpenGL14);
    glPointParameterfv = (void *) MHGLGetProcAddress ("glPointParameterfv", true, &OpenGL14);
    glSecondaryColor3b = (void *) MHGLGetProcAddress ("glSecondaryColor3b", true, &OpenGL14);
    glSecondaryColor3bv = (void *) MHGLGetProcAddress ("glSecondaryColor3bv", true, &OpenGL14);
    glSecondaryColor3d = (void *) MHGLGetProcAddress ("glSecondaryColor3d", true, &OpenGL14);
    glSecondaryColor3dv = (void *) MHGLGetProcAddress ("glSecondaryColor3dv", true, &OpenGL14);
    glSecondaryColor3f = (void *) MHGLGetProcAddress ("glSecondaryColor3f", true, &OpenGL14);
    glSecondaryColor3fv = (void *) MHGLGetProcAddress ("glSecondaryColor3fv", true, &OpenGL14);
    glSecondaryColor3i = (void *) MHGLGetProcAddress ("glSecondaryColor3i", true, &OpenGL14);
    glSecondaryColor3iv = (void *) MHGLGetProcAddress ("glSecondaryColor3iv", true, &OpenGL14);
    glSecondaryColor3s = (void *) MHGLGetProcAddress ("glSecondaryColor3s", true, &OpenGL14);
    glSecondaryColor3sv = (void *) MHGLGetProcAddress ("glSecondaryColor3sv", true, &OpenGL14);
    glSecondaryColor3ub = (void *) MHGLGetProcAddress ("glSecondaryColor3ub", true, &OpenGL14);
    glSecondaryColor3ubv = (void *) MHGLGetProcAddress ("glSecondaryColor3ubv", true, &OpenGL14);
    glSecondaryColor3ui = (void *) MHGLGetProcAddress ("glSecondaryColor3ui", true, &OpenGL14);
    glSecondaryColor3uiv = (void *) MHGLGetProcAddress ("glSecondaryColor3uiv", true, &OpenGL14);
    glSecondaryColor3us = (void *) MHGLGetProcAddress ("glSecondaryColor3us", true, &OpenGL14);
    glSecondaryColor3usv = (void *) MHGLGetProcAddress ("glSecondaryColor3usv", true, &OpenGL14);
    glSecondaryColorPointer = (void *) MHGLGetProcAddress ("glSecondaryColorPointer", true, &OpenGL14);
    glWindowPos2d = (void *) MHGLGetProcAddress ("glWindowPos2d", true, &OpenGL14);
    glWindowPos2dv = (void *) MHGLGetProcAddress ("glWindowPos2dv", true, &OpenGL14);
    glWindowPos2f = (void *) MHGLGetProcAddress ("glWindowPos2f", true, &OpenGL14);
    glWindowPos2fv = (void *) MHGLGetProcAddress ("glWindowPos2fv", true, &OpenGL14);
    glWindowPos2i = (void *) MHGLGetProcAddress ("glWindowPos2i", true, &OpenGL14);
    glWindowPos2iv = (void *) MHGLGetProcAddress ("glWindowPos2iv", true, &OpenGL14);
    glWindowPos2s = (void *) MHGLGetProcAddress ("glWindowPos2s", true, &OpenGL14);
    glWindowPos2sv = (void *) MHGLGetProcAddress ("glWindowPos2sv", true, &OpenGL14);
    glWindowPos3d = (void *) MHGLGetProcAddress ("glWindowPos3d", true, &OpenGL14);
    glWindowPos3dv = (void *) MHGLGetProcAddress ("glWindowPos3dv", true, &OpenGL14);
    glWindowPos3f = (void *) MHGLGetProcAddress ("glWindowPos3f", true, &OpenGL14);
    glWindowPos3fv = (void *) MHGLGetProcAddress ("glWindowPos3fv", true, &OpenGL14);
    glWindowPos3i = (void *) MHGLGetProcAddress ("glWindowPos3i", true, &OpenGL14);
    glWindowPos3iv = (void *) MHGLGetProcAddress ("glWindowPos3iv", true, &OpenGL14);
    glWindowPos3s = (void *) MHGLGetProcAddress ("glWindowPos3s", true, &OpenGL14);
    glWindowPos3sv = (void *) MHGLGetProcAddress ("glWindowPos3sv", true, &OpenGL14);
    glLockArrays = (void *) MHGLGetProcAddress ("glLockArrays", true, &OpenGL14);
    glUnlockArrays = (void *) MHGLGetProcAddress ("glUnlockArrays", true, &OpenGL14);

    // OpenGL 1.5 - get them all but don't crash on a fail cos they're optional
    glGenQueries = (void *) MHGLGetProcAddress ("glGenQueries", true, &OpenGL15);
    glDeleteQueries = (void *) MHGLGetProcAddress ("glDeleteQueries", true, &OpenGL15);
    glIsQuery = (void *) MHGLGetProcAddress ("glIsQuery", true, &OpenGL15);
    glEndQuery = (void *) MHGLGetProcAddress ("glEndQuery", true, &OpenGL15);
    glGetQueryiv = (void *) MHGLGetProcAddress ("glGetQueryiv", true, &OpenGL15);
    glGetQueryObjectiv = (void *) MHGLGetProcAddress ("glGetQueryObjectiv", false, &OpenGL15);
    glGetQueryObjectuiv = (void *) MHGLGetProcAddress ("glGetQueryObjectuiv", true, &OpenGL15);
    glBindBuffer = (void *) MHGLGetProcAddress ("glBindBuffer", true, &OpenGL15);
    glDeleteBuffers = (void *) MHGLGetProcAddress ("glDeleteBuffers", true, &OpenGL15);
    glGenBuffers = (void *) MHGLGetProcAddress ("glGenBuffers", true, &OpenGL15);
    glIsBuffer = (void *) MHGLGetProcAddress ("glIsBuffer", true, &OpenGL15);
    glBufferData = (void *) MHGLGetProcAddress ("glBufferData", true, &OpenGL15);
    glBufferSubData = (void *) MHGLGetProcAddress ("glBufferSubData", true, &OpenGL15);
    glGetBufferSubData = (void *) MHGLGetProcAddress ("glGetBufferSubData", true, &OpenGL15);
    glMapBuffer = (void *) MHGLGetProcAddress ("glMapBuffer", true, &OpenGL15);
    glUnmapBuffer = (void *) MHGLGetProcAddress ("glUnmapBuffer", true, &OpenGL15);
    glGetBufferParameteriv = (void *) MHGLGetProcAddress ("glGetBufferParameteriv", true, &OpenGL15);
    glGetBufferPointerv = (void *) MHGLGetProcAddress ("glGetBufferPointerv", true, &OpenGL15);

    // vertex programs
    glGetVertexAttribdvARB = (void *) MHGLGetProcAddress( "glGetVertexAttribdvARB", true, &OpenGLVert );
    glGetVertexAttribfvARB	= (void *) MHGLGetProcAddress( "glGetVertexAttribfvARB", true, &OpenGLVert );
    glGetVertexAttribivARB	= (void *) MHGLGetProcAddress( "glGetVertexAttribivARB", true, &OpenGLVert );
    glGetVertexAttribPointervARB	= (void *) MHGLGetProcAddress( "glGetVertexAttribPointervARB", true, &OpenGLVert );

    // fragment programs
    glProgramStringARB = (void *) MHGLGetProcAddress( "glProgramStringARB", true, &OpenGLFrag);
    glBindProgramARB = (void *) MHGLGetProcAddress( "glBindProgramARB", true, &OpenGLFrag);
    glDeleteProgramsARB = (void *) MHGLGetProcAddress( "glDeleteProgramsARB", true, &OpenGLFrag);
    glGenProgramsARB = (void *) MHGLGetProcAddress( "glGenProgramsARB", true, &OpenGLFrag);
    glProgramEnvParameter4dARB = (void *) MHGLGetProcAddress( "glProgramEnvParameter4dARB", true, &OpenGLFrag);
    glProgramEnvParameter4dvARB = (void *) MHGLGetProcAddress( "glProgramEnvParameter4dvARB", true, &OpenGLFrag);
    glProgramEnvParameter4fARB = (void *) MHGLGetProcAddress( "glProgramEnvParameter4fARB", true, &OpenGLFrag);
    glProgramEnvParameter4fvARB = (void *) MHGLGetProcAddress( "glProgramEnvParameter4fvARB", true, &OpenGLFrag);
    glProgramLocalParameter4dARB = (void *) MHGLGetProcAddress( "glProgramLocalParameter4dARB", true, &OpenGLFrag);
    glProgramLocalParameter4dvARB = (void *) MHGLGetProcAddress( "glProgramLocalParameter4dvARB", true, &OpenGLFrag);
    glProgramLocalParameter4fARB = (void *) MHGLGetProcAddress( "glProgramLocalParameter4fARB", true, &OpenGLFrag);
    glProgramLocalParameter4fvARB = (void *) MHGLGetProcAddress( "glProgramLocalParameter4fvARB", true, &OpenGLFrag);
    glGetProgramEnvParameterdvARB = (void *) MHGLGetProcAddress( "glGetProgramEnvParameterdvARB", true, &OpenGLFrag);
    glGetProgramEnvParameterfvARB = (void *) MHGLGetProcAddress( "glGetProgramEnvParameterfvARB", true, &OpenGLFrag);
    glGetProgramLocalParameterdvARB = (void *) MHGLGetProcAddress( "glGetProgramLocalParameterdvARB", true, &OpenGLFrag);
    glGetProgramLocalParameterfvARB = (void *) MHGLGetProcAddress( "glGetProgramLocalParameterfvARB", true, &OpenGLFrag);
    glGetProgramivARB = (void *) MHGLGetProcAddress( "glGetProgramivARB", true, &OpenGLFrag);
    glGetProgramStringARB = (void *) MHGLGetProcAddress( "glGetProgramStringARB", true, &OpenGLFrag);

    glGetIntegerv (GL_MAX_TEXTURE_SIZE, &GLCaps.GL_MaxTextureSize);
    Con_Printf ("\nMaximum Texture Size: %i x %i\n", GLCaps.GL_MaxTextureSize, GLCaps.GL_MaxTextureSize);

    if (OpenGL12)
    {
        Con_Printf ("\nOpenGL 1.2 OK\n");
    }
    else
    {
        Con_Printf ("\nOpenGL 1.2 Not fully supported\n");
    }

    if (OpenGL13)
    {
        Con_Printf ("\nOpenGL 1.3 OK\n");
    }
    else
    {
        Con_Printf ("\nOpenGL 1.3 Not fully supported\n");
    }

    if (OpenGL14)
    {
        Con_Printf ("\nOpenGL 1.4 OK\n");
    }
    else
    {
        Con_Printf ("\nOpenGL 1.4 Not fully supported\n");
    }

    if (OpenGL15)
    {
        Con_Printf ("\nOpenGL 1.5 OK\n");
    }
    else
    {
        Con_Printf ("\nOpenGL 1.5 Not fully supported\n");
    }

    if (OpenGLVert)
    {
        Con_Printf ("\nOpenGL Vertex Programs OK\n");
    }
    else
    {
        Con_Printf ("\nOpenGL Vertex Programs Not fully supported\n");
    }

    if (OpenGLFrag)
    {
        Con_Printf ("\nOpenGL Fragment Programs OK\n");
    }
    else
    {
        Con_Printf ("\nOpenGL Fragment Programs Not fully supported\n");
    }

    // check OpenGL 1.5 optional features
    if (GL_State.gl_version >= 1.5)
    {
        GL_ExtensionBits |= HAS_OCCLUSION;
        Con_Printf ("\nFound core feature: occlusion_query\n");
    }
    else
    {
        if (GL_CheckExtension ("occlusion_query", false, false))
        {
            GL_ExtensionBits |= HAS_OCCLUSION;
        }
    }

    if (GL_ExtensionBits & HAS_OCCLUSION)
    {
        glGetQueryiv (GL_SAMPLES_PASSED, GL_QUERY_COUNTER_BITS, &GLCaps.GL_OcclusionBits);
        Con_Printf ("\n    %i Occlusion Query Bits\n", GLCaps.GL_OcclusionBits);
    }
    else
    {
        GLCaps.GL_OcclusionBits = 0;
    }

    // check other non-core and/or proprietary extensions
    if (GL_CheckExtension ("texture_filter_anisotropic", false, false))
    {
        GL_ExtensionBits |= HAS_ANISOTROPIC;

        glGetFloatv (GL_MAX_TEXTURE_MAX_ANISOTROPY, &GLCaps.GL_MaxAnisotropy);
        Con_Printf ("\nMaximum Anisotropic Filter: %i\n", (int) GLCaps.GL_MaxAnisotropy);
    }
    else
    {
        GLCaps.GL_MaxAnisotropy = 1;
    }

    if (GL_CheckExtension ("GL_EXT_compiled_vertex_array", false, true))
    {
        GL_ExtensionBits |= HAS_CVA;
        Con_Printf ("\nFound Extension GL_EXT_compiled_vertex_array\n");
    }

    if (GL_CheckExtension ("GL_ARB_vertex_program", false, false))
    {
        GL_ExtensionBits |= HAS_ARBVERTEXPROGRAMS;
        Con_Printf ("\nFound Extension GL_ARB_vertex_program\n");
    }

    if (GL_CheckExtension ("GL_ARB_fragment_program", false, false))
    {
        GL_ExtensionBits |= HAS_ARBFRAGMENTSHADERS;
        Con_Printf ("\nFound Extension GL_ARB_fragment_program\n");
    }

    // get the number of TMUs available
    Con_Printf("\nFound %i TMUs support.\n", GLCaps.GL_TextureUnits);
    Con_Printf ("\nMinimal Required OpenGL Startup OK\n");
}

/*
===============
GL_SetupState -- johnfitz

does all the stuff from GL_Init that needs to be done every time a new GL render context is created
GL_Init will still do the stuff that only needs to be done once
===============
*/
void GL_SetupState (void)
{
    float	GL_atten[3] = {0.01f, 0.0f, 0.01f};
    int		GL_ColorBitsRed;
    int		GL_ColorBitsGreen;
    int		GL_ColorBitsBlue;
    int		GL_DepthBits;
    int		GL_StencilBits;

    // get some caps
    glGetIntegerv (GL_RED_BITS, &GL_ColorBitsRed);
    glGetIntegerv (GL_GREEN_BITS, &GL_ColorBitsGreen);
    glGetIntegerv (GL_BLUE_BITS, &GL_ColorBitsBlue);
    glGetIntegerv (GL_DEPTH_BITS, &GL_DepthBits);
    glGetIntegerv (GL_STENCIL_BITS, &GL_StencilBits);

    Con_Printf ("\nBuffers\n");
    Con_Printf ("\n   Color:   %2i Bits\n", GL_ColorBitsRed + GL_ColorBitsGreen + GL_ColorBitsBlue);
    Con_Printf ("\n   Depth:   %2i Bits\n", GL_DepthBits);
    Con_Printf ("\n   Stencil: %2i Bits\n", GL_StencilBits);

    GL_GetFunctionPointers ();

    // default states
    glClearColor (0.0, 0.0, 0.0, 1.0);
    glCullFace(GL_FRONT);
    glEnable(GL_TEXTURE_2D);
    glAlphaFunc(GL_GREATER, 0.666f);
    glPolygonOffset(0.05f, 0);

    glPolygonMode (GL_FRONT_AND_BACK, GL_FILL);
    glShadeModel (GL_SMOOTH);

    glTexParameterf (GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR);
    glTexParameterf (GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
    glTexParameterf (GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
    glTexParameterf (GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);

    glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

    glTexEnvf (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);

    // set up the hints for drawing
    glHint (GL_POINT_SMOOTH_HINT, GL_NICEST);
    glHint (GL_LINE_SMOOTH_HINT, GL_NICEST);
    glHint (GL_POLYGON_SMOOTH_HINT, GL_NICEST);
    glHint (GL_FOG_HINT, GL_NICEST);
    glHint (GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);
    glHint (GL_GENERATE_MIPMAP_HINT, GL_NICEST);

    // explicitly enable dithering
    glEnable (GL_DITHER);

    // use fragment depth fog for higher quality
    glFogf (GL_FOG_COORDINATE_SOURCE, GL_FRAGMENT_DEPTH);

    // point parameters
    // ATI doesn't have glPointParameteri...
    glPointParameterf (GL_POINT_SIZE_MIN, 2);
    glPointParameterf (GL_POINT_SIZE_MAX, 40);
    glPointParameterfv (GL_POINT_DISTANCE_ATTENUATION, GL_atten);
    glPointSize (40);
}

/*
===============
GL_Init
===============
*/
void GL_Init (void)
{
    int		i;

    // set baseline gl state
    GL_InitState ();

    GL_State.gl_vendor = (const char *)glGetString (GL_VENDOR);
    Con_Printf ("\nGL_VENDOR:   %s\n", GL_State.gl_vendor);

    GL_State.gl_renderer = (const char *)glGetString (GL_RENDERER);
    Con_Printf ("GL_RENDERER: %s\n", GL_State.gl_renderer);

    GL_State.gl_version = atof ((const char *)glGetString (GL_VERSION));
    Con_Printf ("GL_VERSION:  %0.2f\n", GL_State.gl_version);

    GL_State.gl_extensions = (const char *)glGetString (GL_EXTENSIONS);
    Con_DPrintf ("\nGL_EXTENSIONS: %s\n", GL_State.gl_extensions);

    if (GL_State.gl_version < 1.4)
    {
        Con_Printf ("\nWARNING\nGL_VERSION %f detected - you may experience reduced functionality\n", GL_State.gl_version);
    }

    for (i = 0; i < (int)strlen (GL_State.gl_renderer) - 7; i++)
    {
        // test different ways of reporting this
        if (GL_State.gl_renderer[i] == 'f' || GL_State.gl_renderer[i] == 'F' &&	GL_State.gl_renderer[i + 1] == 'x' || GL_State.gl_renderer[i + 1] == 'X')
        {
            int j;

            if (GL_State.gl_renderer[i + 2] == ' ')
            {
                j = i + 3;
            }
            else
            {
                j = i + 2;
            }

            if (j + 3 >= (int)strlen (GL_State.gl_renderer)) break;

            if (GL_State.gl_renderer[j] == '5')
            {
                GL_State.NP2BigDirtyFib = true;
                Con_Printf ("\nWARNING\nNV 3x Chipset detected - you may experience reduced functionality\n");
                break;
            }
        }
    }
    GL_SetupState ();
}


/*
=================
GL_BeginRendering

=================
*/
void GL_BeginRendering (int *x, int *y, int *width, int *height)
{
    *x = *y = 0;

    *width = WindowRect.right - WindowRect.left;
    *height = WindowRect.bottom - WindowRect.top;
}


void GL_EndRendering (void)
{
    if (!scr_skipupdate)
    {
        SwapBuffers(maindc);
    }

// handle the mouse state when windowed if that's changed
    if (modestate == MS_WINDOWED)
    {
        if (!_windowed_mouse.value)
        {
            if (windowed_mouse)
            {
                IN_DeactivateMouse ();
                IN_ShowMouse ();
                windowed_mouse = false;
            }
        }
        else
        {
            windowed_mouse = true;

            if (key_dest == key_game && !mouseactive && ActiveApp)
            {
                IN_ActivateMouse ();
                IN_HideMouse ();
            }
            else if (mouseactive && key_dest != key_game)
            {
                IN_DeactivateMouse ();
                IN_ShowMouse ();
            }
        }
    }
}

void VID_SetPalette (unsigned char *palette)
{
    byte			*pal;
    unsigned		r, g, b;
    unsigned		v;
    unsigned short	i;
    unsigned		*table;

    pal = palette;
    table = d_8to24table;

    for (i=0 ; i<256 ; i++)
    {
        r = pal[0];
        g = pal[1];
        b = pal[2];
        pal += 3;

        v = (255 << 24) + (r << 0) + (g << 8) + (b << 16);
        *table++ = v;
    }
    d_8to24table[255] &= 0xffffff;	// 255 is transparent
}

void VID_SetDefaultMode (void)
{
    IN_DeactivateMouse ();
}


void VID_Shutdown (void)
{
    HGLRC hRC;
    HDC	  hDC;

    if (vid_initialized)
    {
        vid_canalttab = false;
        hRC = wglGetCurrentContext ();
        hDC = wglGetCurrentDC ();

        wglMakeCurrent (NULL, NULL);

        GL_DeleteAllTextures ();

        if (hRC) wglDeleteContext(hRC);
        if (hDC && dibwindow) ReleaseDC (dibwindow, hDC);
        if (modestate == MS_FULLDIB) ChangeDisplaySettings (NULL, 0);
        if (maindc && dibwindow) ReleaseDC (dibwindow, maindc);

        AppActivate(false, false);
    }
}


//==========================================================================


BOOL bSetupPixelFormat (HDC hDC)
{
    static PIXELFORMATDESCRIPTOR pfd =
    {
        sizeof(PIXELFORMATDESCRIPTOR),	// size of this pfd
        1,								// version number
        PFD_DRAW_TO_WINDOW 				// support window
        |  PFD_SUPPORT_OPENGL 			// support OpenGL
        |  PFD_DOUBLEBUFFER ,			// double buffered
        PFD_TYPE_RGBA,					// RGBA type
        24,								// 24-bit color depth (8 for each of r/g/b/a)
        0, 0, 0, 0, 0, 0,				// color bits ignored
        0,								// 8 bits of your finest alpha if you please, my good man (set to none cos some implementations fall back to sw!!!)
        0,								// shift bit ignored
        0,								// no accumulation buffer
        0, 0, 0, 0, 					// accum bits ignored
        24,								// 24 bit z buffer.  because we have removed gl_ztrick, we can get away with this
        8,								// 8 bit stencil buffer
        0,								// no auxiliary buffer
        PFD_MAIN_PLANE,					// main layer
        0,								// reserved
        0, 0, 0							// layer masks ignored
    };

    int pixelformat;

    if ((pixelformat = ChoosePixelFormat (hDC, &pfd)) == 0)
    {
        MessageBox (NULL, "ChoosePixelFormat failed", "Error", MB_OK);
        return FALSE;
    }

    if (SetPixelFormat (hDC, pixelformat, &pfd) == FALSE)
    {
        MessageBox (NULL, "SetPixelFormat failed", "Error", MB_OK);
        return FALSE;
    }

    // the stencil buffer should be available on everything aside from 3dfx so this is really a redundant
    // test, cos Realm will work on everything aside from a 3dfx...  besides, if you have a 3dfx, not
    // having a stencil buffer will be least of your problems...
    if (!pfd.cStencilBits)
    {
        Sys_Error ("Zomg!: ancient 3dfx card detected\n");
    }
    return TRUE;
}

byte        scantokey[128] =
{
//  0           1       2       3       4       5       6       7
//  8           9       A       B       C       D       E       F
    0  ,    27,     '1',    '2',    '3',    '4',    '5',    '6',
    '7',    '8',    '9',    '0',    '-',    '=',    K_BACKSPACE, 9, // 0
    'q',    'w',    'e',    'r',    't',    'y',    'u',    'i',
    'o',    'p',    '[',    ']',    13 ,    K_CTRL,'a',  's',      // 1
    'd',    'f',    'g',    'h',    'j',    'k',    'l',    ';',
    '\'' ,    '`',    K_SHIFT,'\\',  'z',    'x',    'c',    'v',      // 2
    'b',    'n',    'm',    ',',    '.',    '/',    K_SHIFT,'*',
    K_ALT,' ',   0  ,    K_F1, K_F2, K_F3, K_F4, K_F5,   // 3
    K_F6, K_F7, K_F8, K_F9, K_F10, K_PAUSE  ,    0  , K_HOME,
    K_UPARROW,K_PGUP,'-',K_LEFTARROW,'5',K_RIGHTARROW,'+',K_END, //4
    K_DOWNARROW,K_PGDN,K_INS,K_DEL,0,0,             0,              K_F11,
    K_F12,0  ,    0  ,    0  ,    0  ,    0  ,    0  ,    0,        // 5
    0  ,    0  ,    0  ,    0  ,    0  ,    0  ,    0  ,    0,
    0  ,    0  ,    0  ,    0  ,    0  ,    0  ,    0  ,    0,        // 6
    0  ,    0  ,    0  ,    0  ,    0  ,    0  ,    0  ,    0,
    0  ,    0  ,    0  ,    0  ,    0  ,    0  ,    0  ,    0         // 7
};

byte        shiftscantokey[128] =
{
//  0           1       2       3       4       5       6       7
//  8           9       A       B       C       D       E       F
    0  ,    27,     '!',    '@',    '#',    '$',    '%',    '^',
    '&',    '*',    '(',    ')',    '_',    '+',    K_BACKSPACE, 9, // 0
    'Q',    'W',    'E',    'R',    'T',    'Y',    'U',    'I',
    'O',    'P',    '{',    '}',    13 ,    K_CTRL,'A',  'S',      // 1
    'D',    'F',    'G',    'H',    'J',    'K',    'L',    ':',
    '"' ,    '~',    K_SHIFT,'|',  'Z',    'X',    'C',    'V',      // 2
    'B',    'N',    'M',    '<',    '>',    '?',    K_SHIFT,'*',
    K_ALT,' ',   0  ,    K_F1, K_F2, K_F3, K_F4, K_F5,   // 3
    K_F6, K_F7, K_F8, K_F9, K_F10, K_PAUSE  ,    0  , K_HOME,
    K_UPARROW,K_PGUP,'_',K_LEFTARROW,'%',K_RIGHTARROW,'+',K_END, //4
    K_DOWNARROW,K_PGDN,K_INS,K_DEL,0,0,             0,              K_F11,
    K_F12,0  ,    0  ,    0  ,    0  ,    0  ,    0  ,    0,        // 5
    0  ,    0  ,    0  ,    0  ,    0  ,    0  ,    0  ,    0,
    0  ,    0  ,    0  ,    0  ,    0  ,    0  ,    0  ,    0,        // 6
    0  ,    0  ,    0  ,    0  ,    0  ,    0  ,    0  ,    0,
    0  ,    0  ,    0  ,    0  ,    0  ,    0  ,    0  ,    0         // 7
};


/*
=======
MapKey

Map from windows to quake keynums
=======
*/
int MapKey (int key)
{
    key = (key>>16)&255;
    if (key > 127)
        return 0;
    if (scantokey[key] == 0)
        Con_DPrintf("key 0x%02x has no translation\n", key);
    return scantokey[key];
}

/*
===================================================================

MAIN WINDOW

===================================================================
*/

/*
================
ClearAllStates
================
*/
void ClearAllStates (void)
{
    int		i;

// send an up event for each key, to make sure the server clears them all
    for (i=0 ; i<256 ; i++)
    {
        Key_Event (i, false);
    }
    Key_ClearStates ();
    IN_ClearStates ();
}

void AppActivate(BOOL fActive, BOOL minimize)
/****************************************************************************
*
* Function:     AppActivate
* Parameters:   fActive - True if app is activating
*
* Description:  If the application is activating, then swap the system
*               into SYSPAL_NOSTATIC mode so that our palettes will display
*               correctly.
*
****************************************************************************/
{
    static BOOL	sound_active;

    ActiveApp = fActive;
    Minimized = minimize;

// enable/disable sound on focus gain/loss
    if (!ActiveApp && sound_active)
    {
        S_BlockSound ();
        sound_active = false;
    }
    else if (ActiveApp && !sound_active)
    {
        S_UnblockSound ();
        sound_active = true;
    }

    if (fActive)
    {
        if (modestate == MS_FULLDIB)
        {
            IN_ActivateMouse ();
            IN_HideMouse ();
            if (vid_canalttab && vid_wassuspended)
            {
                vid_wassuspended = false;
                ChangeDisplaySettings (&gdevmode, CDS_FULLSCREEN);
                ShowWindow(mainwindow, SW_SHOWNORMAL);
                MoveWindow (mainwindow, 0, 0, gdevmode.dmPelsWidth, gdevmode.dmPelsHeight, false);
            }
        }
        else if ((modestate == MS_WINDOWED) && _windowed_mouse.value && key_dest == key_game)
        {
            IN_ActivateMouse ();
            IN_HideMouse ();
        }
    }

    if (!fActive)
    {
        if (modestate == MS_FULLDIB)
        {
            IN_DeactivateMouse ();
            IN_ShowMouse ();
            if (vid_canalttab)
            {
                ChangeDisplaySettings (NULL, 0);
                vid_wassuspended = true;
            }
        }
        else if ((modestate == MS_WINDOWED) && _windowed_mouse.value)
        {
            IN_DeactivateMouse ();
            IN_ShowMouse ();
        }
    }
}

#ifndef WM_GRAPHNOTIFY
#define WM_GRAPHNOTIFY  WM_USER + 13
#endif

/* main window procedure */
LONG WINAPI MainWndProc (
    HWND    hWnd,
    UINT    uMsg,
    WPARAM  wParam,
    LPARAM  lParam)
{
    LONG    lRet = 1;
    int		fActive, fMinimized, temp;
    extern unsigned int uiWheelMessage;

    if ( uMsg == uiWheelMessage )
        uMsg = WM_MOUSEWHEEL;

    switch (uMsg)
    {
    case WM_KILLFOCUS:
        if (modestate == MS_FULLDIB)
            ShowWindow(mainwindow, SW_SHOWMINNOACTIVE);
        break;

    case WM_CREATE:
        break;

    case WM_MOVE:
        window_x = (int) LOWORD(lParam);
        window_y = (int) HIWORD(lParam);
        VID_UpdateWindowStatus ();
        break;

    case WM_KEYDOWN:
    case WM_SYSKEYDOWN:
        Key_Event (MapKey(lParam), true);
        break;

    case WM_KEYUP:
    case WM_SYSKEYUP:
        Key_Event (MapKey(lParam), false);
        break;

    case WM_SYSCHAR:
        // keep Alt-Space from happening
        break;

        // this is complicated because Win32 seems to pack multiple mouse events into
        // one update sometimes, so we always check all states and look for events
    case WM_LBUTTONDOWN:
    case WM_LBUTTONUP:
    case WM_RBUTTONDOWN:
    case WM_RBUTTONUP:
    case WM_MBUTTONDOWN:
    case WM_MBUTTONUP:
    case WM_MOUSEMOVE:
        temp = 0;

        if (wParam & MK_LBUTTON)
            temp |= 1;

        if (wParam & MK_RBUTTON)
            temp |= 2;

        if (wParam & MK_MBUTTON)
            temp |= 4;

        IN_MouseEvent (temp);

        break;

        // JACK: This is the mouse wheel with the Intellimouse
        // Its delta is either positive or neg, and we generate the proper
        // Event.
    case WM_MOUSEWHEEL:
        if ((short) HIWORD(wParam) > 0)
        {
            Key_Event(K_MWHEELUP, true);
            Key_Event(K_MWHEELUP, false);
        }
        else
        {
            Key_Event(K_MWHEELDOWN, true);
            Key_Event(K_MWHEELDOWN, false);
        }
        break;

    case WM_SIZE:
        break;

    case WM_CLOSE:
        if (MessageBox (mainwindow, "Are you sure you want to quit?", "Confirm Exit", MB_YESNO | MB_SETFOREGROUND | MB_ICONQUESTION) == IDYES)
        {
            Sys_Quit ();
        }
        break;

    case WM_ACTIVATE:
        fActive = LOWORD(wParam);
        fMinimized = (BOOL) HIWORD(wParam);
        AppActivate(!(fActive == WA_INACTIVE), fMinimized);

        // fix the leftover Alt from any Alt-Tab or the like that switched us away
        ClearAllStates ();

        //R00k re-init the mouse after alt-tabbing please... (v1.9)
        if (!mouseactive)
        {
            IN_Shutdown ();
            IN_StartupMouse();
            IN_ActivateMouse ();
            IN_HideMouse ();
        }
        break;

    case WM_DESTROY:
    {
        if (dibwindow)
        {
            DestroyWindow (dibwindow);
        }
        PostQuitMessage (0);
    }
    break;

    case WM_GRAPHNOTIFY:
        lRet = CDAudio_MessageHandler (hWnd, uMsg, wParam, lParam);
        break;

    default:
        /* pass all unhandled messages to DefWindowProc */
        lRet = DefWindowProc (hWnd, uMsg, wParam, lParam);
        break;
    }

    /* return 1 if handled message, 0 if not */
    return lRet;
}


/*
=================
VID_NumModes
=================
*/
int VID_NumModes (void)
{
    return nummodes;
}


/*
=================
VID_GetModePtr
=================
*/
vmode_t *VID_GetModePtr (int modenum)
{

    if ((modenum >= 0) && (modenum < nummodes))
        return &modelist[modenum];
    else
        return &badmode;
}


/*
=================
VID_GetModeDescription
=================
*/
char *VID_GetModeDescription (int mode)
{
    char		*pinfo;
    vmode_t		*pv;
    static char	temp[100];

    if ((mode < 0) || (mode >= nummodes))
        return NULL;

    if (!leavecurrentmode)
    {
        pv = VID_GetModePtr (mode);
        pinfo = pv->modedesc;
    }
    else
    {
        sprintf (temp, "Desktop resolution (%d x %d x %d %dHz)",
                 modelist[MODE_FULLSCREEN_DEFAULT].width,
                 modelist[MODE_FULLSCREEN_DEFAULT].height,
                 modelist[MODE_FULLSCREEN_DEFAULT].bpp,
                 modelist[MODE_FULLSCREEN_DEFAULT].refreshrate);

        pinfo = temp;
    }
    return pinfo;
}


// KJB: Added this to return the mode driver name in description for console

char *VID_GetExtModeDescription (int mode)
{
    static char	pinfo[40];
    vmode_t		*pv;

    if ((mode < 0) || (mode >= nummodes))
        return NULL;

    pv = VID_GetModePtr (mode);
    if (modelist[mode].type == MS_FULLDIB)
    {
        if (!leavecurrentmode)
        {
            sprintf(pinfo,"%s fullscreen", pv->modedesc);
        }
        else
        {
            sprintf (pinfo, "Desktop resolution (%d x %d x %d %dHz)",
                     modelist[MODE_FULLSCREEN_DEFAULT].width,
                     modelist[MODE_FULLSCREEN_DEFAULT].height,
                     modelist[MODE_FULLSCREEN_DEFAULT].bpp,
                     modelist[MODE_FULLSCREEN_DEFAULT].refreshrate);
        }
    }
    else
    {
        if (modestate == MS_WINDOWED)
            sprintf(pinfo, "%s windowed", pv->modedesc);
        else
            sprintf(pinfo, "windowed");
    }
    return pinfo;
}


/*
=================
VID_DescribeCurrentMode_f
=================
*/
void VID_DescribeCurrentMode_f (void)
{
    Con_Printf ("%s\n", VID_GetExtModeDescription (vid_modenum));
}


/*
=================
VID_NumModes_f
=================
*/
void VID_NumModes_f (void)
{
    if (nummodes == 1)
    {
        Con_Printf ("%d video mode is available\n", nummodes);
    }
    else
    {
        Con_Printf ("%d video modes are available\n", nummodes);
    }
}


/*
=================
VID_DescribeMode_f
=================
*/
void VID_DescribeMode_f (void)
{
    int		t, modenum;

    modenum = atoi (Cmd_Argv(1));

    t = leavecurrentmode;
    leavecurrentmode = 0;

    Con_Printf ("%s\n", VID_GetExtModeDescription (modenum));

    leavecurrentmode = t;
}


/*
=================
VID_DescribeModes_f
=================
*/
void VID_DescribeModes_f (void)
{
    int			i, lnummodes, t;
    char		*pinfo;
    vmode_t		*pv;

    lnummodes = VID_NumModes ();

    t = leavecurrentmode;
    leavecurrentmode = 0;

    for (i = 1; i < lnummodes ; i++)
    {
        pv = VID_GetModePtr (i);
        pinfo = VID_GetExtModeDescription (i);
        Con_Printf ("%2d: %s\n", i, pinfo);
    }
    leavecurrentmode = t;
}

/*
===================
VID_Restart -- johnfitz -- change video modes on the fly
===================
*/
void VID_Restart (void)
{
    HDC         hdc;
    HGLRC		hrc;
    int         i;
    qboolean	mode_changed = false;
    vmode_t     oldmode;

    if (vid_locked)	return;

    //
    // check cvars against current mode
    //
    if (vid_fullscreen.value)
    {
        if (modelist[vid_default].type == MS_WINDOWED)
        {
            mode_changed = true;
        }
        else if (modelist[vid_default].refreshrate != (int)vid_refreshrate.value)
        {
            mode_changed = true;
        }
    }
    else
    {
        if (modelist[vid_default].type != MS_WINDOWED)
        {
            mode_changed = true;
        }
    }

    if (modelist[vid_default].width != (int)vid_width.value || modelist[vid_default].height != (int)vid_height.value)
    {
        mode_changed = true;
    }

    if (mode_changed)
    {
        //
        // decide which mode to set
        //
        oldmode = modelist[vid_default];

        if (vid_fullscreen.value)
        {
            for (i=1; i<nummodes; i++)
            {
                if (modelist[i].width == (int)vid_width.value &&
                        modelist[i].height == (int)vid_height.value &&
                        modelist[i].bpp == (int)vid_bpp &&
                        modelist[i].refreshrate == (int)vid_refreshrate.value)
                {
                    break;
                }
            }

            if (i == nummodes)
            {
                Con_Printf ("%dx%dx%d %dHz is not a valid fullscreen mode\n",
                            (int)vid_width.value,
                            (int)vid_height.value,
                            (int)vid_bpp,
                            (int)vid_refreshrate.value);
                return;
            }
            windowed = false;
            vid_default = i;
        }
        else //not fullscreen
        {
            hdc = GetDC (NULL);
            if (GetDeviceCaps(hdc, RASTERCAPS) & RC_PALETTE)
            {
                Con_Printf ("Can't run windowed on non-RGB desktop\n");
                ReleaseDC (NULL, hdc);
                return;
            }
            ReleaseDC (NULL, hdc);

            if (vid_width.value < 320)
            {
                Con_Printf ("Window width can't be less than 320\n");
                return;
            }

            if (vid_height.value < 200)
            {
                Con_Printf ("Window height can't be less than 200\n");
                return;
            }

            modelist[0].width = (int)vid_width.value;
            modelist[0].height = (int)vid_height.value;
            sprintf (modelist[0].modedesc, "%dx%dx%d %dHz",
                     modelist[0].width,
                     modelist[0].height,
                     modelist[0].bpp,
                     modelist[0].refreshrate);

            windowed = true;
            vid_default = 0;
        }
        //
        // destroy current window
        //
        hrc = wglGetCurrentContext();
        hdc = wglGetCurrentDC();
        wglMakeCurrent(NULL, NULL);

        vid_canalttab = false;

        if (hdc && dibwindow)
        {
            ReleaseDC (dibwindow, hdc);
        }

        if (modestate == MS_FULLDIB)
        {
            ChangeDisplaySettings (NULL, 0);
        }

        if (maindc && dibwindow)
        {
            ReleaseDC (dibwindow, maindc);
        }
        maindc = NULL;

        if (dibwindow)
        {
            DestroyWindow (dibwindow);
        }

        //
        // set new mode
        //
        VID_SetMode (vid_default, host_basepal);
        maindc = GetDC(mainwindow);
        bSetupPixelFormat(maindc);

        // if bpp changes, recreate render context and reload textures
        if (modelist[vid_default].bpp != oldmode.bpp)
        {
            wglDeleteContext (hrc);
            hrc = wglCreateContext (maindc);

            if (!wglMakeCurrent (maindc, hrc))
            {
                Sys_Error ("VID_Restart: wglMakeCurrent failed");
            }
            GL_SetupState ();
        }
        else if (!wglMakeCurrent (maindc, hrc))
        {
            char	szBuf[80];
            LPVOID	lpMsgBuf;
            DWORD	dw = GetLastError();
            FormatMessage(FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM, NULL, dw, MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT), (LPTSTR) &lpMsgBuf, 0, NULL );
            sprintf(szBuf, "VID_Restart: wglMakeCurrent failed with error %d: %s", dw, lpMsgBuf);
            Sys_Error (szBuf);
        }
        vid_canalttab = true;
    }
}

/*
================
VID_Test -- johnfitz -- like vid_restart, but asks for confirmation after switching modes
================
*/
void VID_Test (void)
{
    vmode_t		oldmode;
    qboolean	mode_changed = false;

    if (vid_locked)	return;
    //
    // check cvars against current mode
    //
    if (vid_fullscreen.value)
    {
        if (modelist[vid_default].type == MS_WINDOWED)
        {
            mode_changed = true;
        }
        else if (modelist[vid_default].refreshrate != (int)vid_refreshrate.value)
        {
            mode_changed = true;
        }
    }
    else
    {
        if (modelist[vid_default].type != MS_WINDOWED)
        {
            mode_changed = true;
        }
    }

    if (modelist[vid_default].width != (int)vid_width.value || modelist[vid_default].height != (int)vid_height.value)
    {
        mode_changed = true;
    }

    if (!mode_changed) return;

    //
    // now try the switch
    //
    oldmode = modelist[vid_default];

    VID_Restart ();

    //pop up confirmation dialoge
    if (!SCR_ModalMessage("Would you like to keep this\nvideo mode? (y/n)\n"))
    {
        //revert cvars and mode
        Cvar_Set ("vid_width", va("%i", oldmode.width));
        Cvar_Set ("vid_height", va("%i", oldmode.height));
        Cvar_Set ("vid_bpp", va("%i", oldmode.bpp));
        Cvar_Set ("vid_refreshrate", va("%i", oldmode.refreshrate));
        Cvar_Set ("vid_fullscreen", (oldmode.type == MS_WINDOWED) ? "0" : "1");
        VID_Restart ();
    }
}

/*
================
VID_Unlock -- johnfitz
================
*/
void VID_Unlock (void)
{
    vid_locked = false;

    //sync up cvars in case they were changed during the lock
    Cvar_Set ("vid_width", va("%i", modelist[vid_default].width));
    Cvar_Set ("vid_height", va("%i", modelist[vid_default].height));
    Cvar_Set ("vid_refreshrate", va("%i", modelist[vid_default].refreshrate));
    Cvar_Set ("vid_fullscreen", (windowed) ? "0" : "1");
}

void VID_InitDIB (HINSTANCE hInstance)
{
    DEVMODE			devmode;
    WNDCLASS		wc;
    HDC				hdc;

    // Register the window class
    wc.style         = 0;
    wc.lpfnWndProc   = (WNDPROC) MainWndProc;
    wc.cbClsExtra    = 0;
    wc.cbWndExtra    = 0;
    wc.hInstance     = hInstance;
    wc.hIcon         = 0;
    wc.hCursor       = LoadCursor (NULL, IDC_ARROW);
    wc.hbrBackground = NULL;
    wc.lpszMenuName  = 0;
    wc.lpszClassName = "Realm";

    if (!RegisterClass (&wc)) Sys_Error ("Couldn't register window class");

    // set up mode 0 as the windowed mode
    modelist[0].type = MS_WINDOWED;

    if (COM_CheckParm("-width"))
        modelist[0].width = atoi(com_argv[COM_CheckParm("-width")+1]);
    else modelist[0].width = 640;

    if (modelist[0].width < 640) modelist[0].width = 640;

    if (COM_CheckParm("-height"))
        modelist[0].height= atoi(com_argv[COM_CheckParm("-height")+1]);
    else modelist[0].height = modelist[0].width * 3 / 4;

    if (modelist[0].height < 480) modelist[0].height = 480;

    hdc = GetDC(NULL);
    modelist[0].bpp = desktop_bpp = GetDeviceCaps(hdc, BITSPIXEL);
    ReleaseDC(NULL, hdc);

    if (EnumDisplaySettings (NULL, ENUM_CURRENT_SETTINGS, &devmode))
    {
        modelist[0].refreshrate = devmode.dmDisplayFrequency;
    }
    sprintf (modelist[0].modedesc, "%d x %d x %d %dHz", modelist[0].width, modelist[0].height, modelist[0].bpp, modelist[0].refreshrate);

    modelist[0].modenum = MODE_WINDOWED;
    modelist[0].dib = 1;
    modelist[0].fullscreen = 0;
    modelist[0].halfscreen = 0;

    nummodes = 1;
}


/*
=================
VID_EnumFullDIB

Because VID_InitFullDIB enumerates modes twice - once for 32bpp modes only, and the second time for other modes,
I've moved this to it's own function
=================
*/
void VID_EnumFullDIB (qboolean bpp32only)
{
    DEVMODE	devmode;
    int		modenum;
    int		originalnummodes;
    int		i;
    qboolean existingmode;

    for (modenum = 0, originalnummodes = nummodes; ; modenum++)
    {
        // too many modes!!!  no point in testing any further
        if (nummodes >= MAX_MODE_LIST) break;

        // enumerate the next available mode, break out if we fail to enumerate
        // (per ms-help://MS.MSDNQTR.2005JAN.1033/gdi/devcons_84oj.htm)
        if (!EnumDisplaySettings (NULL, modenum, &devmode)) break;

        // condtion for first run - only report 32 bpp modes (get them to the top of the list so as to ensure that high-res
        // 32 bpp modes have a better chance of being available)
        if (bpp32only && (devmode.dmBitsPerPel != 32)) continue;

        // condtion for second run - report everything except 32bpp modes (get them at the end of the list, so it
        // doesn't matter as much if some of them fall off)
        if (!bpp32only && (devmode.dmBitsPerPel == 32)) continue;

        // invalid mode conditions
        if (devmode.dmBitsPerPel < 16) continue;			// not enough bpp
        if (devmode.dmBitsPerPel > 32) continue;			// too many bpp
        if (devmode.dmPelsWidth > MAXWIDTH) continue;		// too wide
        if (devmode.dmPelsHeight > MAXHEIGHT) continue;		// too high
        if (devmode.dmPelsWidth < 640) continue;			// not wide enough
        if (devmode.dmPelsHeight < 480) continue;			// not high enough

        // set up the fields used by the CDS test
        devmode.dmFields = DM_BITSPERPEL | DM_PELSWIDTH | DM_PELSHEIGHT | DM_DISPLAYFREQUENCY;

        // run the CDS in test mode
        if (ChangeDisplaySettings (&devmode, CDS_TEST | CDS_FULLSCREEN) == DISP_CHANGE_SUCCESSFUL)
        {
            // worked OK so store out the mode
            modelist[nummodes].type = MS_FULLDIB;
            modelist[nummodes].width = devmode.dmPelsWidth;
            modelist[nummodes].height = devmode.dmPelsHeight;
            modelist[nummodes].modenum = 0;
            modelist[nummodes].halfscreen = 0;
            modelist[nummodes].dib = 1;
            modelist[nummodes].fullscreen = 1;
            modelist[nummodes].bpp = devmode.dmBitsPerPel;
            modelist[nummodes].refreshrate = devmode.dmDisplayFrequency;
            sprintf (modelist[nummodes].modedesc, "%d x %d x %d %dHz", devmode.dmPelsWidth,	devmode.dmPelsHeight, devmode.dmBitsPerPel, devmode.dmDisplayFrequency);

            // now see does this mode already exist.  this is needed cos modes can differ in matters other than
            // width, height and bpp (commonly, refresh rate)
            for (i = originalnummodes, existingmode = false; i < nummodes; i++)
            {
                // invalid conditions
                if (modelist[nummodes].width != modelist[i].width) continue;				// different width
                if (modelist[nummodes].height != modelist[i].height) continue;				// different height
                if (modelist[nummodes].bpp != modelist[i].bpp) continue;					// different bpp
                if (modelist[nummodes].refreshrate != modelist[i].refreshrate) continue;	// different refreshrate

                // width, height and bpp all match, so it's something inconsequential like refresh rate that's different
                existingmode = true;
                break;
            }

            // increment the mode counter if it's a new mode
            if (!existingmode)
            {
                nummodes++;
            }
        }
    }
}


/*
=================
VID_InitFullDIB
=================
*/
void VID_InitFullDIB (HINSTANCE hInstance)
{
    // store the original number of modes here
    int originalnummodes = nummodes;

    VID_EnumFullDIB (true);		// find all 32 bpp modes
    VID_EnumFullDIB (false);	// find all non-32 bpp modes

    // this should never happen
    if (nummodes == originalnummodes)
    {
        Con_SafePrintf ("No fullscreen DIB modes found\n");
    }
}

static void VID_CheckGamma (unsigned char *pal)
{
    float			f, inf;
    unsigned char	palette[768];
    int				i;

    if ((i = COM_CheckParm("-gamma")) == 0)
    {
        // no need to send through the calculation
        vid_gamma = 1.0;
        return;
    }
    else
    {
        vid_gamma = atof(com_argv[i+1]);
    }

    for (i=0 ; i<768 ; i++)
    {
        f = pow ((pal[i] + 1) / 256.0, vid_gamma);
        inf = f * 255 + 0.5;
        if (inf < 0) inf = 0;
        if (inf > 255) inf = 255;
        palette[i] = inf;
    }
    memcpy (pal, palette, sizeof(palette));
}

/*
===================
VID_Init
===================
*/
void VID_Init (unsigned char *palette)
{
    int		i;
    int		width, height, bpp = 0, refreshrate;
    HDC		hdc;
    int		desktopmatch = 0;

    Cvar_RegisterVariable (&vid_mode);
    Cvar_RegisterVariable (&vid_wait);
    Cvar_RegisterVariable (&vid_nopageflip);
    Cvar_RegisterVariable (&_vid_wait_override);
    Cvar_RegisterVariable (&_vid_default_mode);
    Cvar_RegisterVariable (&_vid_default_mode_win);
    Cvar_RegisterVariable (&vid_config_x);
    Cvar_RegisterVariable (&vid_config_y);
    Cvar_RegisterVariable (&vid_stretch_by_2);
    Cvar_RegisterVariable (&_windowed_mouse);
    Cvar_RegisterVariable (&vid_fullscreen);
    Cvar_RegisterVariable (&vid_width);
    Cvar_RegisterVariable (&vid_height);
    Cvar_RegisterVariable (&vid_refreshrate);

    Cmd_AddCommand ("vid_unlock", VID_Unlock);
    Cmd_AddCommand ("vid_restart", VID_Restart);
    Cmd_AddCommand ("vid_test", VID_Test);
    Cmd_AddCommand ("vid_nummodes", VID_NumModes_f);
    Cmd_AddCommand ("vid_describecurrentmode", VID_DescribeCurrentMode_f);
    Cmd_AddCommand ("vid_describemode", VID_DescribeMode_f);
    Cmd_AddCommand ("vid_describemodes", VID_DescribeModes_f);

    hIcon = LoadIcon (global_hInstance, MAKEINTRESOURCE (IDI_ICON2));

    InitCommonControls ();

    VID_InitDIB (global_hInstance);
    VID_InitFullDIB (global_hInstance);

    // destroy the dialog here so that it doesn't interfere with any errors
    DestroyWindow (hwnd_dialog);

    // get the current device context
    hdc = GetDC (NULL);

    if (COM_CheckParm ("-window"))
    {
        if (GetDeviceCaps (hdc, RASTERCAPS) & RC_PALETTE)
        {
            Sys_Error ("Can't run in non-RGB mode");
        }
        windowed = true;
        vid_default = MODE_WINDOWED;
    }
    else
    {
        // attempt to match a specified mode with the list of available modes
        if (nummodes == 1) Sys_Error ("No RGB fullscreen modes available");

        windowed = false;

        if (COM_CheckParm ("-mode"))
        {
            vid_default = atoi (com_argv[COM_CheckParm ("-mode") + 1]);
        }
        else
        {
            if (COM_CheckParm ("-current"))
            {
                // take the current desktop resolution and colour depth as the mode, and write these into the
                // default fullscreen mode (mode 1 - originally 640 x 480 x 16)
                modelist[MODE_FULLSCREEN_DEFAULT].width = GetSystemMetrics (SM_CXSCREEN);
                modelist[MODE_FULLSCREEN_DEFAULT].height = GetSystemMetrics (SM_CYSCREEN);
                modelist[MODE_FULLSCREEN_DEFAULT].bpp = GetDeviceCaps (hdc, BITSPIXEL);
                modelist[MODE_FULLSCREEN_DEFAULT].refreshrate = GetDeviceCaps(hdc, VREFRESH);

                vid_default = MODE_FULLSCREEN_DEFAULT;
                leavecurrentmode = 1;
            }
            else
            {
                // check for a specified width or height, default to desktop dimensions if not specified
                if (COM_CheckParm ("-width"))
                {
                    width = atoi (com_argv[COM_CheckParm ("-width") + 1]);

                    // if height is unspecified, pick a matching height at 1.3333 aspect ratio
                    if (COM_CheckParm ("-height"))
                        height = atoi (com_argv[COM_CheckParm ("-height") + 1]);
                    else height = (width * 3) / 4;
                }
                else if (COM_CheckParm ("-height"))
                {
                    // specified height
                    height = atoi (com_argv[COM_CheckParm ("-height") + 1]);

                    // if height is unspecified, pick a matching height at 1.3333 aspect ratio
                    if (COM_CheckParm ("-width"))
                        width = atoi (com_argv[COM_CheckParm ("-width") + 1]);
                    else width = (height * 4) / 3;
                }
                else
                {
                    // unspecified, so use the desktop resolution
                    width = GetSystemMetrics (SM_CXSCREEN);
                    height = GetSystemMetrics (SM_CYSCREEN);

                    // one match to the desktop
                    desktopmatch++;
                }

                // check for a specified bpp, default to desktop bpp if not specified
                if (COM_CheckParm ("-bpp"))
                {
                    bpp = atoi (com_argv[COM_CheckParm ("-bpp") + 1]);
                }
                else
                {
                    // desktop bpp
                    bpp = GetDeviceCaps (hdc, BITSPIXEL);

                    // one match to the desktop
                    desktopmatch++;
                }

                // check for a specified vid_refreshrate.value
                if (COM_CheckParm ("-refreshrate"))
                {
                    refreshrate = atoi (com_argv[COM_CheckParm ("-refreshrate") + 1]);
                }
                else
                {
                    refreshrate = GetDeviceCaps(hdc, VREFRESH);

                    // one match to the desktop
                    desktopmatch++;
                }

                // oh you son of a bi...
                if (desktopmatch == 3)
                {
                    // if both dimensions depth and refresh match the desktop, treat it as if it was -current
                    modelist[MODE_FULLSCREEN_DEFAULT].width = width;
                    modelist[MODE_FULLSCREEN_DEFAULT].height = height;
                    modelist[MODE_FULLSCREEN_DEFAULT].bpp = bpp;
                    modelist[MODE_FULLSCREEN_DEFAULT].refreshrate = refreshrate;

                    vid_default = MODE_FULLSCREEN_DEFAULT;
                    leavecurrentmode = 1;
                }
                else
                {
                    // user specified mode, so check the available modes to see if it's valid
                    for (i = 1, vid_default = 0; i < nummodes; i++)
                    {
                        // try to match the mode
                        if ((modelist[i].width == width) && (modelist[i].height == height) && (modelist[i].bpp == bpp) && (modelist[i].refreshrate == refreshrate))
                        {
                            // found it
                            vid_default = i;
                            break;
                        }
                    }

                    // no suitable video mode found
                    if (!vid_default)
                    {
                        Sys_Error ("Specified video mode %d x %d x %d %dHz not available.", width, height, bpp, refreshrate);
                    }
                }
            }
        }
    }

    // release the device context
    ReleaseDC (NULL, hdc);

    // vid color depth
    vid_bpp = bpp;

    // vid is initialized
    vid_initialized = true;

    // lock these
    vid.conwidth = 640;
    vid.conheight = 480;

    // software quake legacy shite
    vid.colormap = host_colormap;
    vid.fullbright = 256 - LittleLong (*((int *) vid.colormap + 2048));

    VID_CheckGamma (palette);
    VID_SetPalette (palette);

    VID_SetMode (vid_default, palette);

    maindc = GetDC (mainwindow);
    bSetupPixelFormat (maindc);

    // attempt to bring up OpenGL
    baseRC = wglCreateContext (maindc);

    if (!baseRC) Sys_Error ("Could not initialize GL (wglCreateContext failed).\n\n",
                                "Make sure you are in at least 16 bit mode, and try reducing the resolution or running -window.\n\n",
                                "If it still fails, you probably need a new video card...");

    if (!wglMakeCurrent (maindc, baseRC))
    {
        Sys_Error ("wglMakeCurrent failed");
    }
    GL_Init ();

    vid_realmode = vid_modenum;

    vid_menudrawfn = VID_MenuDraw;
    vid_menukeyfn = VID_MenuKey;
    vid_menucmdfn = VID_Menu_f;

    strlcpy_s (badmode.modedesc, "Bad mode");
    vid_canalttab = true;

    if (COM_CheckParm ("-fullsbar"))
    {
        fullsbardraw = true;
    }
    VID_Menu_Init();

    //johnfitz -- command line vid settings should override config file settings.
    //so we have to lock the vid mode from now until after all config files are read.
    if (COM_CheckParm("-width") || COM_CheckParm("-height") || COM_CheckParm("-bpp") || COM_CheckParm("-window"))
    {
        vid_locked = true;
    }
}


//========================================================
// Video menu stuff
//========================================================

/*
================
VID_SyncCvars -- johnfitz -- set vid cvars to match current video mode
================
*/
void VID_SyncCvars (void)
{
    Cvar_Set ("vid_width", va("%i", modelist[vid_default].width));
    Cvar_Set ("vid_height", va("%i", modelist[vid_default].height));
    Cvar_Set ("vid_refreshrate", va("%i", modelist[vid_default].refreshrate));
    Cvar_Set ("vid_fullscreen", (windowed) ? "0" : "1");
}

//==========================================================================
//
//  NEW VIDEO MENU -- johnfitz
//
//==========================================================================

extern void M_Menu_Options_f (void);
extern void M_Print (int cx, int cy, char *str);
extern void M_PrintWhite (int cx, int cy, char *str);
extern void M_DrawCharacter (int cx, int line, int num);
extern void M_DrawTransPic (int x, int y, qpic_t *pic);
extern void M_DrawPic (int x, int y, qpic_t *pic);
extern void M_DrawCheckbox (int x, int y, int on);

extern qboolean   m_entersound;

#define VIDEO_OPTIONS_ITEMS 6
int      video_cursor_table[] = {48, 56, 64, 72, 88, 96};
int      video_options_cursor = 0;

typedef struct
{
    int width,height;
} vid_menu_mode;

//TODO: replace these fixed-length arrays with hunk_allocated buffers
vid_menu_mode vid_menu_modes[MAX_MODE_LIST];
int vid_menu_nummodes = 0;
int vid_menu_numrates = 0;
int vid_menu_rates[20];

/*
================
VID_Menu_Init
================
*/
void VID_Menu_Init (void)
{
    int i, j, h, w;

    for (i=1; i<nummodes; i++) //start i at mode 1 because 0 is windowed mode
    {
        w = modelist[i].width;
        h = modelist[i].height;

        for (j=0; j<vid_menu_nummodes; j++)
        {
            if (vid_menu_modes[j].width == w &&	vid_menu_modes[j].height == h) break;
        }

        if (j == vid_menu_nummodes)
        {
            vid_menu_modes[j].width = w;
            vid_menu_modes[j].height = h;
            vid_menu_nummodes++;
        }
    }
}

/*
================
VID_Menu_RebuildRateList

regenerates rate list based on current vid_width, vid_height and vid_bpp
================
*/
void VID_Menu_RebuildRateList (void)
{
    int i, j, r;

    vid_menu_numrates = 0;

    for (i = 1; i < nummodes; i++) //start i at mode 1 because 0 is windowed mode
    {
        //rate list is limited to rates available with current width/height
        if (modelist[i].width != vid_width.value ||	modelist[i].height != vid_height.value) continue;

        r = modelist[i].refreshrate;

        for (j = 0; j < vid_menu_numrates; j++)
        {
            if (vid_menu_rates[j] == r)	break;
        }

        if (j == vid_menu_numrates)
        {
            vid_menu_rates[j] = r;
            vid_menu_numrates++;
        }
    }

    //if vid_refreshrate is not in the new list, change vid_refreshrate
    for (i = 0; i < vid_menu_numrates; i++)
    {
        if (vid_menu_rates[i] == (int)(vid_refreshrate.value)) break;
    }

    if (i == vid_menu_numrates)
    {
        Cvar_Set ("vid_refreshrate", va ("%i", vid_menu_rates[0]));
    }
}

/*
================
VID_Menu_ChooseNextMode

chooses next resolution in order, then updates vid_width and
vid_height cvars, then updates bpp and refreshrate lists
================
*/
void VID_Menu_ChooseNextMode (int dir)
{
    int i;

    for (i = 0; i < vid_menu_nummodes; i++)
    {
        if (vid_menu_modes[i].width == vid_width.value && vid_menu_modes[i].height == vid_height.value)	break;
    }

    if (i == vid_menu_nummodes) //can't find it in list, so it must be a custom windowed res
    {
        i = 0;
    }
    else
    {
        i += dir;

        if (i >= vid_menu_nummodes)
        {
            i = 0;
        }
        else if (i < 0)
        {
            i = vid_menu_nummodes - 1;
        }
    }
    Cvar_Set ("vid_width",va("%i",vid_menu_modes[i].width));
    Cvar_Set ("vid_height",va("%i",vid_menu_modes[i].height));
    VID_Menu_RebuildRateList ();
}

/*
================
VID_Menu_ChooseNextRate

chooses next refresh rate in order, then updates vid_refreshrate cvar
================
*/
void VID_Menu_ChooseNextRate (int dir)
{
    int i;

    for (i = 0; i < vid_menu_numrates; i++)
    {
        if (vid_menu_rates[i] == vid_refreshrate.value)	break;
    }

    if (i == vid_menu_numrates) //can't find it in list
    {
        i = 0;
    }
    else
    {
        i += dir;

        if (i >= vid_menu_numrates)
        {
            i = 0;
        }
        else if (i < 0)
        {
            i = vid_menu_numrates - 1;
        }
    }
    Cvar_Set ("vid_refreshrate", va("%i", vid_menu_rates[i]));
}

/*
================
VID_MenuKey
================
*/
void VID_MenuKey (int key)
{
    switch (key)
    {
    case K_ESCAPE:
        VID_SyncCvars (); //sync cvars before leaving menu. FIXME: there are other ways to leave menu
        S_LocalSound ("misc/menu1.wav");
        M_Menu_Options_f ();
        break;

    case K_UPARROW:
        S_LocalSound ("misc/menu1.wav");
        video_options_cursor--;
        if (video_options_cursor < 0)
        {
            video_options_cursor = VIDEO_OPTIONS_ITEMS-1;
        }
        break;

    case K_DOWNARROW:
        S_LocalSound ("misc/menu1.wav");
        video_options_cursor++;
        if (video_options_cursor >= VIDEO_OPTIONS_ITEMS)
        {
            video_options_cursor = 0;
        }
        break;

    case K_LEFTARROW:
        S_LocalSound ("misc/menu3.wav");
        switch (video_options_cursor)
        {
        case 0:
            VID_Menu_ChooseNextMode (-1);
            break;
        case 1:
            break;
        case 2:
            VID_Menu_ChooseNextRate (-1);
            break;
        case 3:
            Cbuf_AddText ("toggle vid_fullscreen\n");
            break;
        case 4:
        case 5:
        default:
            break;
        }
        break;

    case K_RIGHTARROW:
        S_LocalSound ("misc/menu3.wav");
        switch (video_options_cursor)
        {
        case 0:
            VID_Menu_ChooseNextMode (1);
            break;
        case 1:
            break;
        case 2:
            VID_Menu_ChooseNextRate (1);
            break;
        case 3:
            if (vid_bpp == desktop_bpp)
            {
                Cbuf_AddText ("toggle vid_fullscreen\n");
            }
            break;
        case 4:
        case 5:
        default:
            break;
        }
        break;

    case K_ENTER:
        m_entersound = true;
        switch (video_options_cursor)
        {
        case 0:
            VID_Menu_ChooseNextMode (1);
            break;
        case 1:
            break;
        case 2:
            VID_Menu_ChooseNextRate (1);
            break;
        case 3:
            if (vid_bpp == desktop_bpp)
            {
                Cbuf_AddText ("toggle vid_fullscreen\n");
            }
            break;
        case 4:
            Cbuf_AddText ("vid_test\n");
            break;
        case 5:
            Cbuf_AddText ("vid_restart\n");
            break;
        default:
            break;
        }
        break;

    default:
        break;
    }
}

/*
================
VID_MenuDraw
================
*/
void VID_MenuDraw (void)
{
    int		i = 0;
    qpic_t	*p;
    char	*title;

    M_DrawTransPic (16, 4, Draw_CachePic ("gfx/qplaque.lmp"));

    p = Draw_CachePic ("gfx/p_option.lmp");
    M_DrawPic ( (320-p->width)/2, 4, p);

    // title
    title = "Video Options";
    M_PrintWhite ((320 - 8 * (int) strlen(title))/2, 32, title);

    // options
    M_Print (16, video_cursor_table[i], "            Video mode");
    M_Print (216, video_cursor_table[i], va("%ix%i", (int)vid_width.value, (int)vid_height.value));
    i++;

    M_Print (16, video_cursor_table[i], "           Color depth");
    M_Print (216, video_cursor_table[i], va("%i [locked]", (int)vid_bpp));
    i++;

    M_Print (16, video_cursor_table[i], "          Refresh rate");
    M_Print (216, video_cursor_table[i], va("%i Hz", (int)vid_refreshrate.value));
    i++;

    M_Print (16, video_cursor_table[i], "            Fullscreen");

    if (vid_bpp == desktop_bpp)
    {
        M_DrawCheckbox (216, video_cursor_table[i], (int)vid_fullscreen.value);
    }
    else
    {
        M_Print (216, video_cursor_table[i], va("%s [locked]", (int)vid_fullscreen.value ? "on" : "off"));
    }
    i++;

    M_Print (16, video_cursor_table[i], "          Test changes");
    i++;

    M_Print (16, video_cursor_table[i], "         Apply changes");

    // cursor
    M_DrawCharacter (200, video_cursor_table[video_options_cursor], 12+((int)(realtime*4)&1));
}

/*
================
VID_Menu_f
================
*/
void VID_Menu_f (void)
{
    key_dest = key_menu;
    m_state = m_video;
    m_entersound = true;

    //set all the cvars to match the current mode when entering the menu
    VID_SyncCvars ();

    //set up bpp and rate lists based on current cvars
    VID_Menu_RebuildRateList ();
}
