/*
Copyright (C) 1996-1997 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/
// mathlib.h

typedef float vec_t;
typedef vec_t vec3_t[3];
typedef vec_t vec5_t[5];

typedef struct
{
    float x;
    float y;
    float z;
} vector_st;

typedef union
{
    vector_st	vec;
    float		vecs[3];
} vector_t;

typedef	int	fixed4_t;
typedef	int	fixed8_t;
typedef	int	fixed16_t;

#ifndef M_PI
#define M_PI		3.14159265358979323846f	// matches value in gcc v2 math.h
#endif

#define RAD2DEG( a ) ( ( (a) * 180.0f ) / M_PI )
#define DEG2RAD( a ) ( a * M_PI ) / 180.0f

struct mplane_s;

extern vec3_t vec3_origin;

#define NANMASK (255<<23)
#define	IS_NAN(x) (((*(int *)&x)&NANMASK) == NANMASK)

#define DotProduct(x,y)(x[0]*y[0]+x[1]*y[1]+x[2]*y[2])
#define CrossProduct(a,b,c)((c)[0]=(a)[1]*(b)[2]-(a)[2]*(b)[1],(c)[1]=(a)[2]*(b)[0]-(a)[0]*(b)[2],(c)[2]=(a)[0]*(b)[1]-(a)[1]*(b)[0])
#define VectorClear(a){a[0]=vec3_origin[0];a[1]=vec3_origin[1];a[2]=vec3_origin[2];}
#define VectorCompare(a,b)(((a)[0]==(b)[0])&&((a)[1]==(b)[1])&&((a)[2]==(b)[2]))
#define VectorMA(a, scale, b, c)((c)[0] = (a)[0] + (scale) * (b)[0],(c)[1] = (a)[1] + (scale) * (b)[1],(c)[2] = (a)[2] + (scale) * (b)[2])

#define PlaneDist(point, plane) ((plane)->type < 3 ? (point)[(plane)->type] : DotProduct((point), (plane)->normal))
#define PlaneDiff(point, plane) (((plane)->type < 3 ? (point)[(plane)->type] : DotProduct((point), (plane)->normal)) - (plane)->dist)

float	anglemod(float a);
float	angledelta (float a);

#define VectorSubtract(a,b,c)								\
(															\
	c[0]	= a[0] - b[0],									\
	c[1]	= a[1] - b[1],									\
	c[2]	= a[2] - b[2]									\
)

#define VectorAdd(a,b,c)									\
(															\
	c[0]	= a[0] + b[0],									\
	c[1]	= a[1] + b[1],									\
	c[2]	= a[2] + b[2]									\
)

#define VectorScale(a,b,c)									\
(															\
	c[0]	= a[0] * b,										\
	c[1]	= a[1] * b,										\
	c[2]	= a[2] * b										\
)

#define VectorCopy(in,out)									\
(															\
	out[0]	= in[0],										\
	out[1]	= in[1],										\
	out[2]	= in[2]											\
)

#define VectorInverse(v)									\
(															\
	v[0]	= -v[0],										\
	v[1]	= -v[1],										\
	v[2]	= -v[2]											\
)

float	Length (vec3_t v);
float   Distance(const vec3_t v, const vec3_t v2);
float   Normalize (vec3_t inout);
void	NormalizeFast (vec3_t v);

int		Q_log2(int val);
void	Q_randinit (void);
int		Q_rand (void);
float	Q_rsqrt (float num);
float	Q_sqrt (float num);

void FloorDivMod (double numer, double denom, int *quotient, int *rem);
fixed16_t Invert24To16(fixed16_t val);
int GreatestCommonDivisor (int i1, int i2);

void	LerpVectors (vec3_t v1, vec_t frac, vec3_t v2, vec3_t v);
void	LerpAngles (vec3_t v1, vec_t frac, vec3_t v2, vec3_t v);
void	VectorAngles (const vec3_t forward, vec3_t angles);
void	AngleVectors (vec3_t angles, vec3_t forward, vec3_t right, vec3_t up);
int		BoxOnPlaneSide (vec3_t emins, vec3_t emaxs, struct mplane_s *plane);
float	anglemod(float a);



#define BOX_ON_PLANE_SIDE(emins, emaxs, p)	\
	(((p)->type < 3)?						\
	(										\
		((p)->dist <= (emins)[(p)->type])?	\
			1								\
		:									\
		(									\
			((p)->dist >= (emaxs)[(p)->type])?\
				2							\
			:								\
				3							\
		)									\
	)										\
	:										\
		BoxOnPlaneSide( (emins), (emaxs), (p)))
