/*
Copyright (C) 1996-1997 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#include "quakedef.h"
#include "winquake.h"
#include "midasdll.h"

MIDASmodule				module;
MIDASmodulePlayHandle	playHandle;

void MIDAS_Error(void)
{
    Con_Printf("Midas Error: %s\n", MIDASgetErrorMessage(MIDASgetLastError()));
}

void MIDAS_Stop(void)
{
    if (sound_midas)
    {
        if (!MIDASstopModule(playHandle))
        {
            MIDAS_Error();
            return;
        }

        if (!MIDASfreeModule(module))
        {
            MIDAS_Error();
            return;
        }

        if (!MIDASclose())
        {
            MIDAS_Error();
            return;
        }
    }
    sound_midas = 0;
}

void MIDAS_Play(void)
{
    filehandle_t	modfile;
    char			modname[256];

    strlcpy(modname, Cmd_Argv(1), sizeof(modname));

    if (sound_midas)
    {
        MIDAS_Stop();
    }

    if (strlen(modname) < 3)
    {
        Con_Print("Format: PlayMod <filename.ext>");
        return;
    }
    COM_FOpenFile (modname, &modfile, FS_READ);

    if (modfile == -1)
    {
        Con_Printf("Unable to find %s\n", modname);
        return;
    }
    Con_Printf("Found %s\n", modname);

    MIDASstartup();

    MIDASsetOption(MIDAS_OPTION_DSOUND_HWND, (DWORD)mainwindow);

    MIDASsetOption(MIDAS_OPTION_DSOUND_MODE, MIDAS_DSOUND_STREAM);

    if (!MIDASinit())
    {
        MIDAS_Error();
        return;
    }

    if ((module = MIDASloadModule(modname)) == NULL)
    {
        MIDAS_Error();
        return;
    }

    if ((playHandle = MIDASplayModule(module, TRUE)) == 0)
    {
        MIDAS_Error();
        return;
    }
    MIDASsetMusicVolume(playHandle, (bgmvolume.value * 64));

    sound_midas = 1;
}

void MIDAS_Init(void)
{
    Cmd_AddCommand ("stopmod", MIDAS_Stop);
    Cmd_AddCommand ("playmod", MIDAS_Play);
}

